"""
History and link-join setting locks for Hangouts.

Config:
    history (dict):
        Mapping from channel names to desired conversation history settings -- ``True`` to keep
        history enabled, ``False`` to keep it disabled.
    linkjoin (dict):
        Mapping from channel names to desired link join settings -- ``True`` to enable joining the
        hangout via link, ``False`` to disable it.
"""

import hangups
from hangups import hangouts_pb2
from voluptuous import ALLOW_EXTRA, Optional, Schema

import immp
from immp.plug.hangouts import HangoutsPlug


HISTORY = {True: hangouts_pb2.OFF_THE_RECORD_STATUS_ON_THE_RECORD,
           False: hangouts_pb2.OFF_THE_RECORD_STATUS_OFF_THE_RECORD}

LINK_JOIN = {True: hangouts_pb2.GROUP_LINK_SHARING_STATUS_ON,
             False: hangouts_pb2.GROUP_LINK_SHARING_STATUS_OFF}


class _Schema:

    config = Schema({Optional("history", default=dict): {str: bool},
                     Optional("linkjoin", default=dict): {str: bool}},
                    extra=ALLOW_EXTRA, required=True)


class HangoutsLockHook(immp.Hook):
    """
    Hook to enforce the history and link-join settings in Hangouts.
    """

    def __init__(self, name, config, host):
        super().__init__(name, _Schema.config(config), host)

    @property
    def channels(self):
        try:
            return {key: {self.host.channels[label]: setting
                          for label, setting in mapping.items()}
                    for key, mapping in self.config.items()}
        except KeyError as e:
            raise immp.HookError("No channel named '{}'".format(e.args[0]))

    async def on_receive(self, sent, source, primary):
        await super().on_receive(sent, source, primary)
        if sent != source or not isinstance(sent.channel.plug, HangoutsPlug):
            return
        conv = sent.channel.plug._convs.get(sent.channel.source)
        if isinstance(sent.raw, hangups.OTREvent):
            setting = HISTORY.get(self.channels["history"].get(sent.channel))
            if setting is None:
                return
            if setting != sent.raw.new_otr_status:
                request = hangouts_pb2.ModifyOTRStatusRequest(
                    request_header=sent.channel.plug._client.get_request_header(),
                    event_request_header=conv._get_event_request_header(),
                    otr_status=setting)
                await sent.channel.plug._client.modify_otr_status(request)
        elif isinstance(sent.raw, hangups.GroupLinkSharingModificationEvent):
            setting = LINK_JOIN.get(self.channels["linkjoin"].get(sent.channel))
            if setting is None:
                return
            if setting != sent.raw.new_status:
                request = hangouts_pb2.SetGroupLinkSharingEnabledRequest(
                    request_header=sent.channel.plug._client.get_request_header(),
                    event_request_header=conv._get_event_request_header(),
                    group_link_sharing_status=setting)
                await sent.channel.plug._client.set_group_link_sharing_enabled(request)



from asyncio import FIRST_COMPLETED, CancelledError, Event, ensure_future, wait
import logging


log = logging.getLogger(__name__)


class PlugStream:
    """
    Message multiplexer, reads messages from multiple asynchronous generators in parallel.

    Instances of this class are async-iterable, producing for each incoming message a tuple
    (:class:`.SentMessage`, :class:`.Message`, :class:`bool`): the physical message received by the
    plug, a source message if originating from within the system, and a primary flag.

    .. warning::
        As per :meth:`.Plug.stream`, only one iterator of this class should be used at once.
    """

    def __init__(self):
        # Mapping from plugs to their stream() generator coroutines.
        self._agens = {}
        # Mapping from coroutine task wrappers back to their tasks.
        self._tasks = {}
        # When a plug is added or removed, the stream wouldn't be able to update until an event
        # arrives.  This is a signal used to recreate the task list without a message.
        self._sync = Event()
        # Generators can't be closed synchronously, schedule the plugs to be done on the next sync.
        self._close = set()

    def add(self, *plugs):
        """
        Connect plugs to the stream.  When the stream is active, their :meth:`.Plug.stream`
        methods will be called to start collecting queued messages.

        Args:
            plugs (.Plug list):
                New plugs to merge in.
        """
        for plug in plugs:
            if plug not in self._agens:
                self._agens[plug] = plug.stream()
        self._sync.set()

    def remove(self, *plugs):
        """
        Disconnect plugs from the stream.  Their :meth:`.Plug.stream` tasks will be cancelled, and
        any last messages will be collected before removing.

        Args:
            plugs (.Plug list):
                Active plugs to remove.
        """
        for plug in plugs:
            if plug in self._agens:
                self._close.add(plug)
        self._sync.set()

    async def _queue(self):
        for plug, coro in self._agens.items():
            if plug not in self._tasks.values():
                log.debug("Queueing receive task for plug %r", plug.name)
                # Poor man's async iteration -- there's no async equivalent to next(gen).
                self._tasks[ensure_future(coro.asend(None))] = plug
        for task, plug in list(self._tasks.items()):
            if plug not in self._agens and plug is not self._sync:
                task.cancel()
                self._close.add(plug)
                del self._tasks[task]
        for plug in self._close:
            log.debug("Cancelling receive task for plug %r", plug.name)
            await self._agens[plug].aclose()
        self._close.clear()
        if self._sync not in self._tasks.values():
            log.debug("Recreating sync task")
            self._sync.clear()
            self._tasks[ensure_future(self._sync.wait())] = self._sync

    async def _stream(self):
        """
        Combined generator to produce messages from all connected plugs as they arrive.  See
        :meth:`.Plug.receive` for more details.  Only a single stream may be active at once.

        Yields:
            (.SentMessage, .Message, bool) tuple:
                Messages received and processed by any connected plug.
        """
        while True:
            log.debug("Waiting for next message")
            try:
                await self._queue()
                done, _ = await wait(self._tasks, return_when=FIRST_COMPLETED)
            except (GeneratorExit, CancelledError):
                for task in self._tasks:
                    task.cancel()
                for coro in self._agens.values():
                    await coro.aclose()
                return
            for task in done:
                plug = self._tasks.pop(task)
                if plug is self._sync:
                    continue
                try:
                    sent, source, primary = task.result()
                except CancelledError:
                    del self._agens[plug]
                except Exception:
                    log.exception("Generator for plug %r exited, recreating", plug.name)
                    self._agens[plug] = plug.stream()
                else:
                    log.info("Received: %r", sent)
                    yield (sent, source, primary)

    def __aiter__(self):
        return self._stream()

    def __repr__(self):
        done = sum(1 for task in self._tasks if task.done())
        pending = len(self._tasks) - done
        return "<{}: {} done, {} pending>".format(self.__class__.__name__, done, pending)



"""
For testing, a plug with no external network.

Messages with sequential identifiers are created and returned every 10 seconds.  Any messages sent
to the plug are cloned and returned as if a network had processed them.
"""

from asyncio import ensure_future, sleep
import logging

import immp


log = logging.getLogger(__name__)


class DummyPlug(immp.Plug):
    """
    Test plug that yields a message every 10 seconds.
    """

    network_name = "Dummy"
    network_id = "dummy"

    def __init__(self, name, config, host):
        super().__init__(name, config, host)
        self.counter = immp.IDGen()
        self.user = immp.User(id="dummy", real_name=name)
        self.channel = immp.Channel(self, "dummy")
        self._task = None

    async def start(self):
        self._task = ensure_future(self._timer())

    async def stop(self):
        if self._task:
            self._task.cancel()
            self._task = None

    async def put(self, channel, msg):
        # Make a clone of the message to echo back out of the generator.
        clone = immp.SentMessage(id=self.counter(),
                                 channel=self.channel,
                                 text=msg.text,
                                 user=msg.user,
                                 action=msg.action)
        log.debug("Returning message: %r", clone)
        self.queue(clone)
        return [clone.id]

    async def _timer(self):
        while True:
            await sleep(10)
            log.debug("Creating next test message")
            self.queue(immp.SentMessage(id=self.counter(),
                                        channel=self.channel,
                                        text="Test",
                                        user=self.user))



"""
Fallback mention and word highlight support for plugs via private channels.

Mentions
~~~~~~~~

Config:
    plugs (str list):
        List of plug names to enable mention alerts for.
    usernames (bool):
        Whether to match network usernames (``True`` by default).
    real-names (bool):
        Whether to match user's display names (``False`` by default).
    ambiguous (bool):
        Whether to notify multiple potential users of an ambiguous mention (``False`` by default).

For networks that don't provide native user mentions, this plug can send users a private message
when mentioned by their username or real name.

A mention is matched from each ``@`` sign until whitespace is encountered.  For real names, spaces
and special characters are ignored, so that e.g. ``@fredbloggs`` will match *Fred Bloggs*.

Partial mentions are supported, failing any exact matches, by basic prefix search on real names.
For example, ``@fred`` will match *Frederick*, and ``@fredb`` will match *Fred Bloggs*.

Subscriptions
~~~~~~~~~~~~~

Config:
    plugs (str list):
        List of plug names to enable subscription alerts for.

Commands:
    sub-add <text>:
        Add a subscription to your trigger list.
    sub-remove <text>:
        Remove a subscription from your trigger list.
    sub-exclude <text>:
        Don't trigger a specific subscription in the current public channel.
    sub-list:
        Show all active subscriptions.

Allows users to opt in to private message notifications when chosen highlight words are used in a
group conversation.

.. note::
    This hook requires an active :class:`.DatabaseHook` to store data.
"""

from asyncio import wait
from collections import defaultdict
import logging
import re

from peewee import CharField, ForeignKeyField
from voluptuous import ALLOW_EXTRA, Optional, Schema

import immp
from immp.hook.command import CommandScope, command
from immp.hook.database import BaseModel, DatabaseHook
from immp.hook.sync import SyncPlug


log = logging.getLogger(__name__)


CROSS = "\N{CROSS MARK}"
TICK = "\N{WHITE HEAVY CHECK MARK}"


class _Schema:

    config = Schema({"groups": [str]}, extra=ALLOW_EXTRA, required=True)

    config_mentions = config.extend({Optional("usernames", default=True): bool,
                                     Optional("real-names", default=False): bool,
                                     Optional("ambiguous", default=False): bool})


class _Skip(Exception):
    # Message isn't applicable to the hook.
    pass


class SubTrigger(BaseModel):
    """
    Individual subscription trigger phrase for an individual user.

    Attributes:
        network (str):
            Network identifier that the user belongs to.
        user (str):
            User identifier as given by the plug.
        text (str):
            Subscription text that they wish to be notified on.
    """

    network = CharField()
    user = CharField()
    text = CharField()

    def __repr__(self):
        return "<{}: #{} {} ({} @ {})>".format(self.__class__.__name__, self.id, repr(self.text),
                                               repr(self.user), repr(self.network))


class SubExclude(BaseModel):
    """
    Exclusion for a trigger in a specific channel.

    Attributes:
        trigger (.SubTrigger):
            Containing trigger instance.
        network (str):
            Network identifier that the channel belongs to.
        user (str):
            Channel's own identifier.
    """

    trigger = ForeignKeyField(model=SubTrigger, related_name="excludes")
    network = CharField()
    channel = CharField()

    def __repr__(self):
        return "<{}: #{} {} @ {} {}>".format(self.__class__.__name__, self.id, repr(self.network),
                                             repr(self.channel), repr(self.trigger))


class _AlertHookBase(immp.Hook):

    group = immp.Group.MergedProperty("groups")

    async def _get_members(self, msg):
        # Sync integration: avoid duplicate notifications inside and outside a synced channel.
        # Commands and excludes should apply to the sync, but notifications are based on the
        # network-native channel.
        if isinstance(msg.channel.plug, SyncPlug):
            # We're in the sync channel, so we've already handled this event in native channels.
            log.debug("Ignoring sync channel: %r", msg.channel)
            raise _Skip
        channel = msg.channel
        synced = SyncPlug.any_sync(self.host, msg.channel)
        if synced:
            # We're in the native channel of a sync, use this channel for reading config.
            log.debug("Translating sync channel: %r -> %r", msg.channel, synced)
            channel = synced
        members = [user for user in (await msg.channel.members()) or []
                   if self.group.has_plug(user.plug)]
        if not members:
            raise _Skip
        return channel, members


class MentionsHook(_AlertHookBase):
    """
    Hook to send mention alerts via private channels.
    """

    def __init__(self, name, config, host):
        super().__init__(name, _Schema.config_mentions(config), host)

    @staticmethod
    def _clean(text):
        return re.sub(r"\W", "", text).lower() if text else None

    def match(self, mention, members):
        """
        Identify users relevant to a mention.

        Args:
            mention (str):
                Raw mention text, e.g. ``@fred``.
            members (.User list):
                List of members in the channel where the mention took place.

        Returns:
            .User set:
                All applicable members to be notified.
        """
        name = self._clean(mention)
        real_matches = set()
        real_partials = set()
        for member in members:
            if self.config["usernames"] and self._clean(member.username) == name:
                # Assume usernames are unique, only match the corresponding user.
                return {member}
            if self.config["real-names"]:
                real = self._clean(member.real_name)
                if real == name:
                    real_matches.add(member)
                if real.startswith(name):
                    real_partials.add(member)
        if real_matches:
            # Assume multiple identical real names is unlikely.
            # If it's the same person with two users, they both get mentioned.
            return real_matches
        elif len(real_partials) == 1 or self.config["ambiguous"]:
            # Return a single partial match if it exists.
            # Only allow multiple partials if enabled, else ignore the mention.
            return real_partials
        else:
            return set()

    async def on_receive(self, sent, source, primary):
        await super().on_receive(sent, source, primary)
        if not primary or not source.text or await sent.channel.is_private():
            return
        try:
            _, members = await self._get_members(sent)
        except _Skip:
            return
        mentioned = set()
        for mention in re.findall(r"@\S+", str(source.text)):
            matches = self.match(mention, members)
            if matches:
                log.debug("Mention %r applies: %r", mention, matches)
                mentioned.update(matches)
            else:
                log.debug("Mention %r doesn't apply", mention)
        for segment in source.text:
            if segment.mention and segment.mention in members:
                log.debug("Segment mention %r applies: %r", segment.text, segment.mention)
                mentioned.add(segment.mention)
        if not mentioned:
            return
        text = immp.RichText()
        if source.user:
            text.append(immp.Segment(source.user.real_name or source.user.username, bold=True),
                        immp.Segment(" mentioned you"))
        else:
            text.append(immp.Segment("You were mentioned"))
        title = await sent.channel.title()
        link = await sent.channel.link()
        if title:
            text.append(immp.Segment(" in "),
                        immp.Segment(title, italic=True))
        text.append(immp.Segment(":\n"))
        text += source.text
        if source.user and source.user.link:
            text.append(immp.Segment("\n"),
                        immp.Segment("Go to user", link=source.user.link))
        if link:
            text.append(immp.Segment("\n"),
                        immp.Segment("Go to channel", link=link))
        tasks = []
        for member in mentioned:
            if member == source.user:
                continue
            private = await sent.channel.plug.channel_for_user(member)
            if private:
                tasks.append(private.send(immp.Message(text=text)))
        if tasks:
            await wait(tasks)


class SubscriptionsHook(_AlertHookBase):
    """
    Hook to send trigger word alerts via private channels.
    """

    def __init__(self, name, config, host):
        super().__init__(name, _Schema.config(config), host)
        self.db = None

    @classmethod
    def _clean(cls, text):
        return re.sub(r"[^\w ]", "", text).lower()

    async def start(self):
        self.db = self.host.resources[DatabaseHook].db
        self.db.create_tables([SubTrigger, SubExclude], safe=True)

    def _test(self, channel, user):
        return self.group.has_plug(channel.plug)

    @command("sub-add", scope=CommandScope.private, test=_test)
    async def add(self, msg, *words):
        """
        Add a subscription to your trigger list.
        """
        text = re.sub(r"[^\w ]", "", " ".join(words)).lower()
        _, created = SubTrigger.get_or_create(network=msg.channel.plug.network_id,
                                              user=msg.user.id, text=text)
        resp = "{} {}".format(TICK, "Subscribed" if created else "Already subscribed")
        await msg.channel.send(immp.Message(text=resp))

    @command("sub-remove", scope=CommandScope.private, test=_test)
    async def remove(self, msg, *words):
        """
        Remove a subscription from your trigger list.
        """
        text = re.sub(r"[^\w ]", "", " ".join(words)).lower()
        count = SubTrigger.delete().where(SubTrigger.network == msg.channel.plug.network_id,
                                          SubTrigger.user == msg.user.id,
                                          SubTrigger.text == text).execute()
        resp = "{} {}".format(TICK, "Unsubscribed" if count else "Not subscribed")
        await msg.channel.send(immp.Message(text=resp))

    @command("sub-list", scope=CommandScope.private, test=_test)
    async def list(self, msg):
        """
        Show all active subscriptions.
        """
        subs = SubTrigger.select().where(SubTrigger.network == msg.user.plug.network_id,
                                         SubTrigger.user == msg.user.id).order_by(SubTrigger.text)
        if subs:
            text = immp.RichText([immp.Segment("Your subscriptions:", bold=True)])
            for sub in subs:
                text.append(immp.Segment("\n- {}".format(sub.text)))
        else:
            text = "No active subscriptions."
        await msg.channel.send(immp.Message(text=text))

    @command("sub-exclude", scope=CommandScope.shared, test=_test)
    async def exclude(self, msg, *words):
        """
        Don't trigger a specific subscription in the current channel.
        """
        text = re.sub(r"[^\w ]", "", " ".join(words)).lower()
        try:
            trigger = SubTrigger.get(network=msg.user.plug.network_id,
                                     user=msg.user.id, text=text)
        except SubTrigger.DoesNotExist:
            resp = "{} Not subscribed".format(CROSS)
        else:
            exclude, created = SubExclude.get_or_create(trigger=trigger,
                                                        network=msg.channel.plug.network_id,
                                                        channel=msg.channel.source)
            if not created:
                exclude.delete_instance()
            resp = "{} {}".format(TICK, "Excluded" if created else "No longer excluded")
        await msg.channel.send(immp.Message(text=resp))

    @staticmethod
    def match(text, channel, present):
        """
        Identify users subscribed to text snippets in a message.

        Args:
            text (str):
                Cleaned message text.
            channel (.Channel):
                Channel where the subscriptions were triggered.
            present (((str, str), .User) dict):
                Mapping from network/user IDs to members of the source channel.

        Returns:
            (.User, str set) dict:
                Mapping from applicable users to their filtered triggers.
        """
        subs = set()
        for sub in SubTrigger.select():
            key = (sub.network, sub.user)
            if key in present and sub.text in text:
                subs.add(sub)
        triggered = defaultdict(set)
        excludes = set(SubExclude.select().where(SubExclude.trigger << subs,
                                                 SubExclude.network == channel.plug.network_id,
                                                 SubExclude.channel == channel.source))
        for trigger in subs - set(exclude.trigger for exclude in excludes):
            triggered[present[(trigger.network, trigger.user)]].add(trigger.text)
        return triggered

    async def channel_migrate(self, old, new):
        count = (SubExclude.update(network=new.plug.network_id, channel=new.source)
                           .where(SubExclude.network == old.plug.network_id,
                                  SubExclude.channel == old.source).execute())
        return count > 0

    async def on_receive(self, sent, source, primary):
        await super().on_receive(sent, source, primary)
        if not primary or not source.text or await sent.channel.is_private():
            return
        try:
            lookup, members = await self._get_members(sent)
        except _Skip:
            return
        present = {(member.plug.network_id, str(member.id)): member for member in members}
        triggered = self.match(self._clean(str(source.text)), lookup, present)
        if not triggered:
            return
        tasks = []
        for member, triggers in triggered.items():
            if member == source.user:
                continue
            private = await member.private_channel()
            if not private:
                continue
            text = immp.RichText()
            mentioned = immp.Segment(", ".join(sorted(triggers)), italic=True)
            if source.user:
                text.append(immp.Segment(source.user.real_name or source.user.username, bold=True),
                            immp.Segment(" mentioned "), mentioned)
            else:
                text.append(mentioned, immp.Segment(" mentioned"))
            title = await sent.channel.title()
            link = await sent.channel.link()
            if title:
                text.append(immp.Segment(" in "),
                            immp.Segment(title, italic=True))
            text.append(immp.Segment(":\n"))
            text += source.text
            if source.user and source.user.link:
                text.append(immp.Segment("\n"),
                            immp.Segment("Go to user", link=source.user.link))
            if link:
                text.append(immp.Segment("\n"),
                            immp.Segment("Go to channel", link=link))
            tasks.append(private.send(immp.Message(text=text)))
        if tasks:
            await wait(tasks)



from asyncio import CancelledError, ensure_future, gather, wait
from itertools import chain
import logging

from .channel import Channel, Group
from .error import ConfigError
from .hook import Hook, ResourceHook
from .plug import Plug
from .stream import PlugStream
from .util import OpenState, pretty_str


log = logging.getLogger(__name__)


class HostGetter:
    """
    Filter property used to return a subset of objects by type.
    """

    def __init__(self, cls):
        self._cls = cls

    def __get__(self, instance, owner):
        if instance:
            return {name: obj for name, obj in instance._objects.items()
                    if isinstance(obj, self._cls)}
        else:
            return self


@pretty_str
class Host:
    """
    Main class responsible for starting, stopping, and interacting with plugs.

    To run as the main coroutine for an application, use :meth:`run` in conjunction with
    :meth:`.AbstractEventLoop.run_until_complete`.  To stop a running host in an async context,
    await :meth:`quit`.

    For finer control, you can call :meth:`open`, :meth:`process` and :meth:`close` explicitly.

    Attributes:
        plugs ((str, .Plug) dict):
            Collection of all registered plug instances, keyed by name.
        hooks ((str, .Hook) dict):
            Collection of all registered hooks, keyed by name.
        plain_hooks ((str, .Hook) dict):
            As above, but excluding resources.
        resources ((class, .ResourceHook) dict):
            Collection of all registered resource hooks, keyed by class.
        running (bool):
            Whether messages from plugs are being processed by the host.
    """

    def __init__(self):
        self._objects = {}
        self.resources = {}
        self._priority = {}
        self._loaded = False
        self._stream = self._process = None

    plugs = HostGetter(Plug)
    channels = HostGetter(Channel)
    groups = HostGetter(Group)
    hooks = HostGetter(Hook)

    @property
    def plain_hooks(self):
        return {name: hook for name, hook in self.hooks.items()
                if not isinstance(hook, ResourceHook)}

    def ordered_hooks(self):
        """
        Sort all registered hooks by priority.

        Returns:
            (.Hook tuple, .Hook tuple) tuple:
                Prioritised and unsorted hooks.
        """
        ordered = tuple(hook for _, hook in sorted(self._priority.items())
                        if hook.state == OpenState.active)
        rest = tuple(hook for hook in chain(self.resources.values(), self.plain_hooks.values())
                     if hook not in ordered and hook.state == OpenState.active)
        return ordered, rest

    def __getitem__(self, key):
        if isinstance(key, str):
            return self._objects[key]
        elif isinstance(key, type) and issubclass(key, ResourceHook):
            return self.resources[key]
        else:
            raise TypeError(key)

    @property
    def running(self):
        # This is the "public" status that external code can query.
        return self._stream is not None

    def add_plug(self, plug):
        """
        Register a plug to the host.

        If the host is already running, use :meth:`join_plug` instead, which will start the plug
        and begin processing messages straight away.

        Args:
            plug (.Plug):
                Existing plug instance to add.

        Returns:
            str:
                Name used to reference this plug.
        """
        if not isinstance(plug, Plug):
            raise TypeError(plug)
        elif plug.name in self._objects:
            raise ConfigError("Plug name '{}' already registered".format(plug.name))
        log.info("Adding plug: %r (%s)", plug.name, plug.__class__.__name__)
        self._objects[plug.name] = plug
        if self._loaded:
            plug.on_load()
        if self._stream:
            self._stream.add(plug)
        return plug.name

    def remove_plug(self, name):
        """
        Unregister an existing plug.

        .. warning::
            This will not notify any hooks with a reference to this plug, nor will it attempt to
            remove it from their state.

        Args:
            name (str):
                Name of a previously registered plug instance to remove.

        Returns:
            .Plug:
                Removed plug instance.
        """
        if not isinstance(self._objects.get(name), Plug):
            raise RuntimeError("Plug '{}' not registered to host".format(name))
        log.info("Removing plug: %s", name)
        plug = self._objects.pop(name)
        for label, channel in list(self.channels.items()):
            if channel.plug == plug:
                self.remove_channel(label)
        if self._stream:
            self._stream.remove(plug)
        return plug

    def add_channel(self, name, channel):
        """
        Register a channel to the host.  The channel's plug must be registered first.

        Args:
            name (str):
                Unique identifier for this channel to be referenced by plugs and hooks.
            channel (.Channel):
                Existing channel instance to add.
        """
        if not isinstance(channel, Channel):
            raise TypeError(channel)
        elif name in self._objects:
            raise ConfigError("Channel name '{}' already registered".format(name))
        log.info("Adding channel: %r (%s/%s)", name, channel.plug.name, channel.source)
        self._objects[name] = channel

    def remove_channel(self, name):
        """
        Unregister an existing channel.

        Args:
            name (str):
                Name of a previously registered channel instance to remove.
        """
        if not isinstance(self._objects.get(name), Channel):
            raise RuntimeError("Channel '{}' not registered to host".format(name))
        log.info("Removing channel: %s", name)
        return self._objects.pop(name)

    def add_group(self, group):
        """
        Register a group to the host.

        Args:
            group (.Group):
                Existing group instance to add.

        Returns:
            str:
                Name used to reference this group.
        """
        if not isinstance(group, Group):
            raise TypeError(group)
        elif group.name in self._objects:
            raise ConfigError("Group name '{}' already registered".format(group.name))
        log.info("Adding group: %s", group.name)
        self._objects[group.name] = group
        return group.name

    def remove_group(self, name):
        """
        Unregister an existing group.

        .. warning::
            This will not notify any hooks with a reference to this group, nor will it attempt to
            remove it from their state.

        Args:
            name (str):
                Name of a previously registered group instance to remove.

        Returns:
            .Group:
                Removed group instance.
        """
        if not isinstance(self._objects.get(name), Group):
            raise RuntimeError("Group '{}' not registered to host".format(name))
        log.info("Removing group: %s", name)
        return self._objects.pop(name)

    def add_hook(self, hook, priority=None):
        """
        Register a hook to the host.

        Args:
            hook (.Hook):
                Existing hook instance to add.
            priority (int):
                Optional ordering constraint, applied to send and receive events.  Hooks registered
                with a priority will be processed first in serial.  Those without will execute
                afterwards, in parallel (in the case of on-receive) or in registration order (for
                before-send and before-receive).

        Returns:
            str:
                Name used to reference this hook.
        """
        if not isinstance(hook, Hook):
            raise TypeError(hook)
        elif hook.name in self._objects:
            raise ConfigError("Hook name '{}' already registered".format(hook.name))
        elif hook.__class__ in self.resources:
            raise ConfigError("Resource class '{}' already registered"
                              .format(hook.__class__.__name__))
        elif priority is not None and priority in self._priority:
            raise ConfigError("Priority {} already registered".format(priority))
        log.info("Adding hook: %r (%s)", hook.name, hook.__class__.__name__)
        self._objects[hook.name] = hook
        if isinstance(hook, ResourceHook):
            log.info("Adding resource: %r (%s)", hook.name, hook.__class__.__name__)
            self.resources[hook.__class__] = hook
        if priority is not None:
            self._priority[priority] = hook
        if self._loaded:
            hook.on_load()
        return hook.name

    def remove_hook(self, name):
        """
        Unregister an existing hook.

        .. warning::
            This will not notify any dependent hooks with a reference to this hook, nor will it
            attempt to remove it from their state.

        Args:
            name (str):
                Name of a previously registered hook instance to remove.

        Returns:
            .Hook:
                Removed hook instance.
        """
        for priority, hook in self._priority.items():
            if hook.name == name:
                del self._priority[priority]
                break
        if not isinstance(self._objects.get(name), Hook):
            raise RuntimeError("Hook '{}' not registered to host".format(name))
        log.info("Removing hook: %s", name)
        hook = self._objects.pop(name)
        if isinstance(hook, ResourceHook):
            log.info("Removing resource: %r (%s)", name, hook.__class__.__name__)
            del self.resources[hook.__class__]
        return hook

    def loaded(self):
        """
        Trigger the on-load event for all plugs and hooks.
        """
        for hook in self.resources.values():
            hook.on_load()
        for plug in self.plugs.values():
            plug.on_load()
        for hook in self.plain_hooks.values():
            hook.on_load()
        self._loaded = True

    async def open(self):
        """
        Connect all open plugs and start all hooks.
        """
        if not self._loaded:
            raise RuntimeError("On-load event must be sent before opening")
        if self.resources:
            await wait([hook.open() for hook in self.resources.values()])
        if self.plugs:
            await wait([plug.open() for plug in self.plugs.values()])
        if self.plain_hooks:
            await wait([hook.open() for hook in self.plain_hooks.values()])

    async def close(self):
        """
        Disconnect all open plugs and stop all hooks.
        """
        if self.plain_hooks:
            await wait([hook.close() for hook in self.plain_hooks.values()])
        if self.plugs:
            await wait([plug.close() for plug in self.plugs.values()])
        if self.resources:
            await wait([hook.close() for hook in self.resources.values()])

    async def _safe_receive(self, hook, sent, source, primary):
        try:
            await hook.on_receive(sent, source, primary)
        except Exception:
            log.exception("Hook %r failed on-receive event", hook.name)

    async def _callback(self, sent, source, primary):
        ordered, rest = self.ordered_hooks()
        for hook in ordered + rest:
            try:
                result = await hook.before_receive(sent, source, primary)
            except Exception:
                log.exception("Hook %r failed before-receive event", hook.name)
                continue
            if result:
                sent = result
            else:
                # Message has been suppressed by a hook.
                return
        for hook in ordered:
            await self._safe_receive(hook, sent, source, primary)
        if rest:
            await gather(*(self._safe_receive(hook, sent, source, primary) for hook in rest))

    async def channel_migrate(self, old, new):
        """
        Issue a migration call to all hooks.

        Args:
            old (.Channel):
                Existing channel with local data.
            new (.Channel):
                Target replacement channel to migrate data to.

        Returns:
            str list:
                Names of hooks that migrated any data for the requested channel.
        """
        hooks = list(self.hooks.values())
        if not hooks:
            return []
        results = await gather(*(hook.channel_migrate(old, new) for hook in hooks))
        return [hook.name for hook, result in zip(hooks, results) if result]

    async def process(self):
        """
        Retrieve messages from plugs, and distribute them to hooks.
        """
        if self._stream:
            raise RuntimeError("Host is already processing")
        self._stream = PlugStream()
        self._stream.add(*self.plugs.values())
        try:
            async for sent, source, primary in self._stream:
                await self._callback(sent, source, primary)
        except StopAsyncIteration:
            log.debug("Plug stream finished")
        finally:
            self._stream = None

    async def run(self):
        """
        Main entry point for running as a full application.  Opens all plugs and hooks, blocks
        (when awaited) for the duration of :meth:`process`, and closes openables during shutdown.
        """
        if self._process:
            raise RuntimeError("Host is already running")
        if not self._loaded:
            self.loaded()
        await self.open()
        try:
            self._process = ensure_future(self.process())
            await self._process
        except CancelledError:
            log.debug("Host run cancelled")
        finally:
            self._process = None
            await self.close()

    async def quit(self):
        """
        Request the running host to stop processing.  This only works if started via :meth:`run`.
        """
        if self._process:
            self._process.cancel()

    def __repr__(self):
        return "<{}: {}P {}C {}G {}R {}H{}>".format(self.__class__.__name__, len(self.plugs),
                                                    len(self.channels), len(self.groups),
                                                    len(self.resources), len(self.plain_hooks),
                                                    " running" if self.running else "")



from asyncio import BoundedSemaphore, Queue
import logging

from .error import PlugError
from .util import Openable, OpenState, pretty_str


log = logging.getLogger(__name__)


@pretty_str
class Plug(Openable):
    """
    Base of all plug classes, handles communication with an external network by converting
    outside data into standardised message objects, and pushing new messages into the network.

    Instantiation may raise :class:`.ConfigError` if the provided configuration is invalid.

    Attributes:
        name (str):
            User-provided, unique name of the plug, used for config references.
        config (dict):
            Reference to the user-provided configuration.
        host (.Host):
            Controlling host instance, providing access to plugs.
        virtual (bool):
            ``True`` if managed by another component (e.g. a hook that exposes plug functionality).
        network_name (str):
            Readable name of the underlying network, for use when displaying info about this plug.
        network_id (str):
            Unique and stable identifier for this plug.

            This should usually vary by the account or keys being used to connect, but persistent
            with later connections.  If a network provides multiple distinct spaces, this should
            also vary by space.
    """

    network_name = network_id = None

    def __init__(self, name, config, host, virtual=False):
        super().__init__()
        self.name = name
        self.config = config
        self.host = host
        self.virtual = virtual
        # Active generator created from get(), referenced to cancel on disconnect.
        self._getter = None
        # Message queue, to move processing from the event stream to the generator.
        self._queue = Queue()
        # Message history, to match up received messages with their sent sources.
        # Mapping from (channel, message ID) to (source message, all IDs).
        self._sent = {}
        # Hook lock, to put a hold on retrieving messages whilst a send is in progress.
        self._lock = BoundedSemaphore()

    async def start(self):
        """
        Start a connection to the external network.

        If using an event-driven framework that yields and runs in the background, you should use
        a signal of some form (e.g. :class:`asyncio.Condition`) to block this method until the
        underlying network is ready for use.
        """
        await super().start()

    async def stop(self):
        """
        Terminate the external network connection.

        Like with :meth:`start`, this should block if needed, such that this method ends when the
        plug can be started again.
        """
        await super().stop()

    async def user_from_id(self, id):
        """
        Retrieve a :class:`.User` based on the underlying network's identifier.

        Args:
            id:
                Network identifier of the user.

        Returns:
            .User:
                Corresponding user instance.
        """
        return None

    async def user_from_username(self, username):
        """
        Retrieve a :class:`.User` based on the underlying network's username.

        Args:
            username:
                Network username of the user.

        Returns:
            .User:
                Corresponding user instance.
        """
        return None

    async def user_is_system(self, user):
        """
        Check if a given user is automated by the plug (for example a bot user from which the plug
        operates).  Hooks may exclude system users from certain operations.

        Returns:
            bool:
                ``True`` if the user relates to the plug itself.
        """
        return False

    async def public_channels(self):
        """
        Retrieve all shared channels known to this plug, either public or otherwise accessible.
        May return ``None`` if the network doesn't support channel discovery.

        Returns:
            .Channel list:
                All available non-private channels.
        """
        return None

    async def private_channels(self):
        """
        Retrieve all private (one-to-one) channels known to this plug.  May return ``None`` if the
        network doesn't support channel discovery.

        Returns:
            .Channel list:
                All available private channels.
        """
        return None

    async def channel_for_user(self, user):
        """
        Retrieve a :class:`.Channel` representing a private (one-to-one) conversation between a
        given user and the service.  Returns ``None`` if the user does not have a private channel.

        Args:
            user (.User):
                Requested user instance.

        Returns:
            .Channel:
                Private channel for this user.
        """
        return None

    async def channel_is_private(self, channel):
        """
        Test if a given channel represents a private (one-to-one) conversation between a given user
        and the service.  May return ``None`` if the network doesn't have a notion of public/shared
        and private channels.

        Args:
            channel (.Channel):
                Requested channel instance.

        Returns:
            bool:
                ``True`` if the channel is private.
        """
        return None

    async def channel_title(self, channel):
        """
        Retrieve the friendly name of this channel, as used in the underlying network.  May return
        ``None`` if the service doesn't have a notion of titles.

        Returns:
            str:
                Display name for the channel.
        """
        return None

    async def channel_link(self, channel):
        """
        Return a URL that acts as a direct link to the given channel.  This is not a join link,
        rather one that opens a conversation in the client (it may e.g. use a custom protocol).

        Returns:
            str:
                Internal deep link to this channel.
        """
        return None

    async def channel_rename(self, channel, title):
        """
        Update the friendly name of this conversation.

        Args:
            title (str):
                New display name for the channel.
        """

    async def channel_members(self, channel):
        """
        Retrieve a :class:`.User` list representing all members of the given channel.  May return
        ``None`` if the plug doesn't recognise the channel, or is unable to query members.

        Args:
            channel (.Channel):
                Requested channel instance.

        Returns:
            .User list:
                Members present in the channel.
        """
        return None

    async def channel_invite(self, channel, user):
        """
        Add the given user to the channel's list of members.

        Args:
            channel (.Channel):
                Requested channel instance.
            user (.User):
                New user to invite.
        """

    async def channel_remove(self, channel, user):
        """
        Remove the given user from the channel's list of members.

        Args:
            channel (.Channel):
                Requested channel instance.
            user (.User):
                Existing user to kick.
        """

    async def channel_history(self, channel, before=None):
        """
        Retrieve the most recent messages sent or received in the given channel.  May return an
        empty list if the plug is unable to query history.

        Args:
            channel (.Channel):
                Requested channel instance.
            before (.Receipt):
                Starting point message, or ``None`` to fetch the most recent.

        Returns:
            .Receipt list:
                Messages from the channel, oldest first.
        """
        return []

    def queue(self, sent):
        """
        Add a new message to the queue, picked up from :meth:`get` by default.

        Args:
            sent (.SentMessage):
                Message received and processed by the plug.
        """
        self._queue.put_nowait(sent)

    def _lookup(self, sent):
        try:
            # If this message was sent by us, retrieve the canonical version.  For source
            # messages split into multiple parts, only make the first raw message primary.
            source, ids = self._sent[(sent.channel, sent.id)]
            primary = ids.index(sent.id) == 0
        except KeyError:
            # The message came from elsewhere, consider it the canonical copy.
            source = sent
            primary = True
        return (sent, source, primary)

    async def stream(self):
        """
        Wrapper method to receive messages from the network.  Plugs should implement their own
        retrieval of messages, scheduled as a background task or managed by another async client,
        then call :meth:`queue` for each received message to have it yielded here.

        This method is called by the :class:`.PlugStream`, in parallel with other plugs to produce
        a single stream of messages across all sources.

        Yields:
            (.SentMessage, .Message, bool) tuple:
                Messages received and processed by the plug, paired with a source message if one
                exists (from a call to :meth:`send`).

        .. warning::
            Because the message buffer is backed by an :class:`asyncio.Queue`, only one generator
            from this method should be used at once -- each message will only be retrieved from the
            queue once, by the first instance that asks for it.
        """
        try:
            while True:
                sent = await self._queue.get()
                async with self._lock:
                    # No critical section here, just wait for any pending messages to be sent.
                    pass
                yield self._lookup(sent)
        except GeneratorExit:
            # Caused by gen.aclose(), in this state we can't yield any further messages.
            log.debug("Immediate exit from plug %r getter", self.name)
        except Exception:
            if not self._queue.empty():
                log.debug("Retrieving queued messages for %r", self.name)
            while not self._queue.empty():
                yield self._lookup(self._queue.get_nowait())

    async def send(self, channel, msg):
        """
        Wrapper method to send a message to the network.  Plugs should implement :meth:`put`
        to convert the framework message into a native representation and submit it.

        Args:
            channel (.Channel):
                Target channel for the new message.
            msg (.Message):
                Original message received from another channel or plug.

        Returns:
            list:
                IDs of new messages sent to the plug.
        """
        if not self.state == OpenState.active:
            raise PlugError("Can't send messages when not active")
        # Allow any hooks to modify the outgoing message before sending.
        original = channel
        ordered, rest = self.host.ordered_hooks()
        for hook in ordered + rest:
            try:
                result = await hook.before_send(channel, msg)
            except Exception:
                log.exception("Hook %r failed before-send event", hook.name)
                continue
            if result:
                channel, msg = result
            else:
                # Message has been suppressed by a hook.
                return []
        if not original == channel:
            log.debug("Redirecting message to new channel: %r", channel)
            # Restart sending with the new channel, including a new round of before-send.
            return await channel.send(msg)
        # When sending messages asynchronously, the network will likely return the new message
        # before the send request returns with confirmation.  Use the lock when sending in order
        # return the new message ID(s) in advance of them appearing in the receive queue.
        async with self._lock:
            ids = await self.put(channel, msg)
        for id in ids:
            self._sent[(channel, id)] = (msg, ids)
        return ids

    async def put(self, channel, msg):
        """
        Take a :class:`.Message` object, and push it to the underlying network.

        Because some plugs may not support combinations of message components (such as text
        and an accompanying image), this method may send more than one physical message.

        Args:
            channel (.Channel):
                Target channel for the new message.
            msg (.Message):
                Original message received from another channel or plug.

        Returns:
            list:
                IDs of new messages sent to the plug.
        """
        return []

    async def delete(self, sent):
        """
        Request deletion of this message, if supported by the network.

        Args:
            sent (.SentMessage):
                Existing message to be removed.
        """

    def on_load(self):
        """
        Perform any additional one-time setup that requires other plugs or hooks to be loaded.
        """

    def __repr__(self):
        return "<{}{}>".format(self.__class__.__name__,
                               ": {}".format(self.network_id) if self.network_id else "")



"""
Identity protocol backbone, and a generic user lookup command.

Config:
    identities (int):
        List of identity provider names from which to allow lookups.
    public (bool):
        ``True`` to allow anyone with access to the ``who`` command to do a lookup, without
        necessarily being identified themselves.

Commands:
    who <name>:
        Recall a known identity and all of its links.

This module defines a subclass for all hooks providing identity services -- no hook is needed from
here if using an identity hook elsewhere.  The :class:`.WhoIsHook` provides a command for users to
query basic identity information.
"""

from asyncio import gather
import logging

from voluptuous import ALLOW_EXTRA, Optional, Schema

import immp
from immp.hook.command import command


CROSS = "\N{CROSS MARK}"
TICK = "\N{WHITE HEAVY CHECK MARK}"


log = logging.getLogger(__name__)


class _Schema:

    config = Schema({"identities": [str],
                     Optional("public", default=False): bool},
                    extra=ALLOW_EXTRA, required=True)


@immp.pretty_str
class Identity:
    """
    Basic representation of an external identity.

    Attributes:
        name (str):
            Common name used across any linked platforms.
        provider (.IdentityProvider):
            Service hook where the identity information was acquired from.
        links (User list):
            Physical platform users assigned to this identity.
        roles (str list):
            Optional set of role names, if applicable to the backend.
    """

    @classmethod
    async def gather(cls, *tasks):
        """
        Helper for fetching users from plugs, filtering out calls with no matches::

            >>> await Identity.gather(plug1.user_from_id(id1), plug2.user_from_id(id2))
            [<Plug1User: '123' 'User'>]

        Args:
            tasks (coroutine list):
                Non-awaited coroutines or tasks.

        Returns:
            list:
                Gathered results of those tasks.
        """
        tasks = list(filter(None, tasks))
        if not tasks:
            return []
        users = []
        for result in await gather(*tasks, return_exceptions=True):
            if isinstance(result, BaseException):
                log.warning("Failed to retrieve user for identity", exc_info=result)
            else:
                users.append(result)
        return users

    def __init__(self, name, provider=None, links=(), roles=()):
        self.name = name
        self.provider = provider
        self.links = links
        self.roles = roles

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and
                (self.name, self.provider) == (other.name, other.provider))

    def __hash__(self):
        return hash(self.name, self.provider)

    def __repr__(self):
        return "<{}: {} x{}{}>".format(self.__class__.__name__, repr(self.name), len(self.links),
                                       " ({})".format(" ".join(self.roles)) if self.roles else "")


class IdentityProvider:
    """
    Interface for hooks to provide identity information from a backing source.
    """

    async def identity_from_name(self, name):
        """
        Look up an identity by the external provider's username for them.

        Args:
            name (str):
                External name to query.

        Returns:
            .Identity:
                Matching identity from the provider, or ``None`` if not found.
        """
        raise NotImplementedError

    async def identity_from_user(self, user):
        """
        Look up an identity by a linked network user.

        Args:
            user (.User):
                Plug user referenced by the identity.

        Returns:
            .Identity:
                Matching identity from the provider, or ``None`` if not found.
        """
        raise NotImplementedError


class WhoIsHook(immp.Hook):
    """
    Hook to provide generic lookup of user profiles across one or more identity providers.
    """

    _identities = immp.ConfigProperty([IdentityProvider])

    def __init__(self, name, config, host):
        super().__init__(name, _Schema.config(config), host)

    @command("who")
    async def who(self, msg, name):
        """
        Recall a known identity and all of its links.
        """
        if self.config["public"]:
            providers = self._identities
        else:
            tasks = (provider.identity_from_user(msg.user) for provider in self._identities)
            providers = [identity.provider for identity in await gather(*tasks) if identity]
        if providers:
            identities = list(filter(None, await gather(*(provider.identity_from_name(name)
                                                          for provider in providers))))
            links = {link for identity in identities for link in identity.links}
            if links:
                text = immp.RichText([immp.Segment(name, bold=True),
                                      immp.Segment(" may appear as:")])
                for user in sorted(links, key=lambda user: user.plug.network_name):
                    text.append(immp.Segment("\n"))
                    text.append(immp.Segment("({}) ".format(user.plug.network_name)))
                    if user.link:
                        text.append(immp.Segment(user.real_name or user.username, link=user.link))
                    elif user.real_name and user.username:
                        text.append(immp.Segment("{} [{}]".format(user.real_name, user.username)))
                    else:
                        text.append(immp.Segment(user.real_name or user.username))
            else:
                text = "{} Name not in use".format(CROSS)
        else:
            text = "{} Not identified".format(CROSS)
        await msg.channel.send(immp.Message(text=text))



"""
Simple custom commands to send preconfigured text messages to channels.

Config:
    commands ((str, str) dict):
        Mapping from command name to rich response text.
"""

from voluptuous import ALLOW_EXTRA, Schema

import immp
from immp.hook.command import command, DynamicCommands


class _Schema:

    config = Schema({"commands": {str: str}}, extra=ALLOW_EXTRA, required=True)


class TextCommandHook(immp.Hook, DynamicCommands):
    """
    Command provider to send configured text responses.
    """

    def __init__(self, name, config, host):
        super().__init__(name, _Schema.config(config), host)

    def commands(self):
        return {self._response.complete(name, name) for name in self.config["commands"]}

    @command()
    async def _response(self, name, msg):
        text = immp.RichText.unraw(self.config["commands"][name], self.host)
        await msg.channel.send(immp.Message(text=text))



"""
Listen for incoming GitHub webhooks.

Config:
    route (str):
        Path to expose the webhook request handler.
    secret (str):
        Shared string between the plug and GitHub's servers.  Optional but recommended, must be
        configured to match the webhook on GitHub.

Go to your repository > Settings > Webhooks > Add webhook, set the URL to match the configured
route, and choose the events you wish to handle.  Message summaries for each event will be emitted
on channels matching the full name of each repository (e.g. ``user/repo``).

.. note::
    This plug requires an active :class:`.WebHook` to receive incoming messages.
"""

import hashlib
import hmac
import logging

from aiohttp import web
from voluptuous import ALLOW_EXTRA, Any, Optional, Schema

import immp
from immp.hook.web import WebHook


log = logging.getLogger(__name__)


class _Schema:

    config = Schema({"route": str,
                     Optional("secret", default=None): Any(str, None)},
                    extra=ALLOW_EXTRA, required=True)

    _sender = {"id": int,
               "login": str,
               "avatar_url": str,
               "html_url": str}

    event = Schema({"repository": {"full_name": str},
                    "sender": _sender},
                   extra=ALLOW_EXTRA, required=True)


class GitHubMessage(immp.Message):
    """
    Repository event originating from GitHub.
    """

    @classmethod
    def from_event(cls, github, type, id, event):
        """
        Convert a `GitHub webhook <https://developer.github.com/webhooks/>`_ payload to a
        :class:`.Message`.

        Args:
            github (.GitHubPlug):
                Related plug instance that provides the event.
            type (str):
                Event type name from the ``X-GitHub-Event`` header.
            id (str):
                GUID of the event delivery from the ``X-GitHub-Delivery`` header.
            event (dict):
                GitHub webhook payload.

        Returns:
            .GitHubMessage:
                Parsed message object.
        """
        repo = event["repository"]["full_name"]
        channel = immp.Channel(github, repo)
        sender = event["sender"]
        user = immp.User(id=sender["id"],
                         plug=github,
                         username=sender["login"],
                         avatar=sender["avatar_url"],
                         link=sender["html_url"],
                         raw=sender)
        text = None
        if type == "push":
            count = len(event["commits"])
            desc = "{} commits".format(count) if count > 1 else event["after"][:7]
            ref = event["ref"].split("/")[1:]
            target = "/".join(ref[1:])
            if ref[0] == "tags":
                action, join = "tagged", "as"
            elif ref[0] == "heads":
                action, join = "pushed", "to"
            else:
                raise NotImplementedError
            text = immp.RichText([immp.Segment("{} ".format(action)),
                                  immp.Segment(desc, link=event["compare"]),
                                  immp.Segment(" {} {} {}".format(join, repo, target))])
            for commit in event["commits"]:
                text.append(immp.Segment("\n* "),
                            immp.Segment(commit["id"][:7], code=True),
                            immp.Segment(" - {}".format(commit["message"])))
        elif type == "release":
            release = event["release"]
            desc = ("{} ({} {})".format(release["name"], repo, release["tag_name"])
                    if release["name"] else release["tag_name"])
            text = immp.RichText([immp.Segment("{} release ".format(event["action"])),
                                  immp.Segment(desc, link=release["html_url"])])
        elif type == "issues":
            issue = event["issue"]
            desc = "{} ({}#{})".format(issue["title"], repo, issue["number"])
            text = immp.RichText([immp.Segment("{} issue ".format(event["action"])),
                                  immp.Segment(desc, link=issue["html_url"])])
        elif type == "pull_request":
            pull = event["pull_request"]
            desc = "{} ({}#{})".format(pull["title"], repo, pull["number"])
            text = immp.RichText([immp.Segment("{} pull request ".format(event["action"])),
                                  immp.Segment(desc, link=pull["html_url"])])
        elif type == "project":
            project = event["project"]
            desc = "{} ({}#{})".format(project["name"], repo, project["number"])
            text = immp.RichText([immp.Segment("{} project ".format(event["action"])),
                                  immp.Segment(desc, link=project["html_url"])])
        elif type == "project_card":
            card = event["project_card"]
            text = immp.RichText([immp.Segment("{} ".format(event["action"])),
                                  immp.Segment("card", link=card["html_url"]),
                                  immp.Segment(" in project:\n"),
                                  immp.Segment(card["note"])])
        elif type == "gollum":
            text = immp.RichText()
            for i, page in enumerate(event["pages"]):
                if i:
                    text.append(immp.Segment(", "))
                text.append(immp.Segment("{} {} wiki page ".format(page["action"], repo)),
                            immp.Segment(page["title"], link=page["html_url"]))
        elif type == "fork":
            fork = event["forkee"]
            text = immp.RichText([immp.Segment("forked {} to ".format(repo)),
                                  immp.Segment(fork["full_name"], link=fork["html_url"])])
        elif type == "watch":
            text = immp.RichText([immp.Segment("starred {}".format(repo))])
        if text:
            return immp.SentMessage(id=id,
                                    channel=channel,
                                    text=text,
                                    user=user,
                                    action=True,
                                    raw=event)
        else:
            raise NotImplementedError


class GitHubPlug(immp.Plug):
    """
    Plug for incoming `GitHub <https://github.com>`_ notifications.
    """

    network_name = "GitHub"
    network_id = "github"

    def __init__(self, name, config, host):
        super().__init__(name, _Schema.config(config), host)
        self._session = None

    def on_load(self):
        log.debug("Registering webhook route")
        self.ctx = self.host.resources[WebHook].context(self.config["route"], __name__)
        self.ctx.route("POST", "", self.handle)

    async def channel_title(self, channel):
        return channel.source

    async def handle(self, request):
        if self.config["secret"]:
            try:
                body = await request.read()
            except ValueError:
                raise web.HTTPBadRequest
            try:
                alg, sig = request.headers["X-Hub-Signature"].split("=", 1)
            except (KeyError, ValueError):
                log.warning("No signature on event, secret needs configuring on webhook")
                raise web.HTTPUnauthorized
            match = hmac.new(self.config["secret"].encode("utf-8"), body, hashlib.sha1).hexdigest()
            if alg != "sha1" or sig != match:
                log.warning("Bad signature on event")
                raise web.HTTPUnauthorized
        try:
            type = request.headers["X-GitHub-Event"]
            id = request.headers["X-GitHub-Delivery"]
            event = _Schema.event(await request.json())
        except (KeyError, ValueError):
            raise web.HTTPBadRequest
        if type == "ping":
            log.debug("Received ping event for %s", event["repository"]["full_name"])
        else:
            try:
                self.queue(GitHubMessage.from_event(self, type, id, event))
            except NotImplementedError:
                log.debug("Ignoring unrecognised event type %r", type)
        return web.Response()



"""
Connect to `Google Hangouts <https://hangouts.google.com>`_ as a regular user.

Config:
    cookie (str):
        Path to a cookie text file read/written by :func:`hangups.get_auth_stdin`.
    read (bool):
        ``True`` (default) to update conversation watermarks on every message, thus marking each
        message as read once processed.

The cookie file is used by hangups to store the access token.  If the file doesn't exist or is
invalid, you will be prompted for Google account credentials at startup.  You can generate a new
cookie file manually by interacting with hangups directly::

    $ python -m hangups.auth

This will generate a cookie file called *refresh_token.txt* in the current directory.

.. note::
    This plug requires the `hangups <https://hangups.readthedocs.io>`_ Python module.
"""

from asyncio import CancelledError, Condition, ensure_future, gather, sleep
from copy import copy
from io import BytesIO
import logging
import re
from urllib.parse import unquote

import hangups
from hangups import hangouts_pb2
from voluptuous import ALLOW_EXTRA, Optional, Schema

import immp


log = logging.getLogger(__name__)


class _Schema:

    config = Schema({"cookie": str,
                     Optional("read", default=True): bool},
                    extra=ALLOW_EXTRA, required=True)


class HangoutsUser(immp.User):
    """
    User present in Hangouts.
    """

    @classmethod
    def from_user(cls, hangouts, user):
        """
        Convert a :class:`hangups.user.User` into a :class:`.User`.

        Args:
            hangouts (.HangoutsPlug):
                Related plug instance that provides the user.
            user (hangups.user.User):
                Hangups user object retrieved from the user list.

        Returns:
            .HangoutsUser:
                Parsed user object.
        """
        # No usernames here, just the ID.
        avatar = re.sub("^//", "https://", user.photo_url) if user.photo_url else None
        return cls(id=user.id_.chat_id,
                   plug=hangouts,
                   real_name=user.full_name,
                   avatar=avatar,
                   raw=user)

    @classmethod
    def from_entity(cls, hangouts, entity):
        """
        Convert a :class:`hangups.hangouts_pb2.Entity` into a :class:`.User`.

        Args:
            hangouts (.HangoutsPlug):
                Related plug instance that provides the user.
            user (hangups.hangouts_pb2.Entity):
                Hangouts entity response retrieved from a
                :class:`hangups.hangouts_pb2.GetEntityByIdRequest`.

        Returns:
            .HangoutsUser:
                Parsed user object.
        """
        avatar = (re.sub("^//", "https://", entity.properties.photo_url)
                  if entity.properties.photo_url else None)
        return cls(id=entity.id_.chat_id,
                   plug=hangouts,
                   real_name=entity.properties.display_name,
                   avatar=avatar,
                   raw=entity)

    @property
    def link(self):
        if self.id:
            return "https://hangouts.google.com/chat/person/{}".format(self.id)


class HangoutsSegment(immp.Segment):
    """
    Plug-friendly representation of Hangouts message formatting.
    """

    @classmethod
    def from_segment(cls, segment):
        """
        Convert a :class:`hangups.ChatMessageSegment` into a :class:`.Segment`.

        Args:
            segment (hangups.ChatMessageSegment):
                Hangups message segment from the conversation event.

        Returns:
            .HangoutsSegment:
                Parsed segment object.
        """
        if segment.type_ == hangouts_pb2.SEGMENT_TYPE_LINE_BREAK:
            return cls("\n")
        link = segment.link_target
        if segment.link_target:
            match = re.match(r"https://www.google.com/url\?(?:[^&]+&)*q=([^&]+)",
                             segment.link_target)
            if match:
                # Strip Google link redirect.
                link = unquote(match[1])
        return cls(segment.text,
                   bold=segment.is_bold,
                   italic=segment.is_italic,
                   underline=segment.is_underline,
                   strike=segment.is_strikethrough,
                   link=link)

    @classmethod
    def _make_segment(cls, text, segment):
        text = re.sub("^ ", "\N{NBSP}", text).replace("  ", " \N{NBSP}")
        return hangups.ChatMessageSegment(text,
                                          (hangouts_pb2.SEGMENT_TYPE_LINK
                                           if segment.link else
                                           hangouts_pb2.SEGMENT_TYPE_TEXT),
                                          is_bold=segment.bold,
                                          is_italic=segment.italic,
                                          is_underline=segment.underline,
                                          is_strikethrough=segment.strike,
                                          link_target=segment.link)

    @classmethod
    def to_segments(cls, segment):
        """
        Convert a :class:`.Segment` into one or more :class:`hangups.ChatMessageSegment` instances.
        This will also attempt to create missing link segments for plain URLs within the text.

        Args:
            segment (.Segment)
                Message segment created by another plug.

        Returns:
            hangups.ChatMessageSegment list:
                Unparsed segment objects.
        """
        unlinked = copy(segment)
        unlinked.link = None
        if segment.link:
            if segment.text == segment.link:
                return [cls._make_segment(segment.text, segment)]
            else:
                return [cls._make_segment("{} [".format(segment.text), unlinked),
                        cls._make_segment(segment.link, segment),
                        cls._make_segment("]", unlinked)]
        segments = []
        for i, line in enumerate(segment.text.split("\n")):
            if i:
                segments.append(hangups.ChatMessageSegment(
                    "\n", hangouts_pb2.SEGMENT_TYPE_LINE_BREAK))
            while True:
                match = re.search(r"\b([a-z-]+://\S+)\b", line)
                if not match:
                    break
                left, right = match.span()
                linked = copy(segment)
                linked.link = line[left:right]
                segments += [cls._make_segment(line[:left], segment),
                             cls._make_segment(linked.link, linked)]
                line = line[right:]
            segments.append(cls._make_segment(line, unlinked))
        return [segment for segment in segments if segment.text]


class HangoutsLocation(immp.Location):

    @classmethod
    def from_embed(cls, embed):
        """
        Convert a :class:`hangouts_pb2.EmbedItem` into a :class:`.Location`.

        Args:
            embed (hangups.hangouts_pb2.EmbedItem):
                Location inside a Hangups message event.

        Returns:
            .HangoutsLocation:
                Parsed location object.
        """
        latitude = longitude = name = address = None
        if embed.place:
            # Chains of elements may be defined but empty at the end.
            if all((embed.place.geo,
                    embed.place.geo.geo_coordinates,
                    embed.place.geo.geo_coordinates.latitude,
                    embed.place.geo.geo_coordinates.longitude)):
                latitude = embed.place.geo.geo_coordinates.latitude
                longitude = embed.place.geo.geo_coordinates.longitude
            if embed.place.name:
                name = embed.place.name
            if all((embed.place.address,
                    embed.place.address.postal_address,
                    embed.place.address.postal_address.street_address)):
                address = embed.place.address.postal_address.street_address
        return cls(latitude=latitude,
                   longitude=longitude,
                   name=name,
                   address=address)

    @classmethod
    def to_place(cls, location):
        """
        Convert a :class:`.Location` into a :class:`hangouts_pb2.EmbedItem`.

        Args:
            location (.Location):
                Location created by another plug.

        Returns:
            hangups.hangouts_pb2.EmbedItem:
                Formatted embed item, suitable for sending inside a
                :class:`hangups.hangouts_pb2.Place` object.
        """
        address = hangouts_pb2.EmbedItem.PostalAddress(street_address=location.address)
        geo = hangouts_pb2.EmbedItem.GeoCoordinates(latitude=location.latitude,
                                                    longitude=location.longitude)
        image_url = ("https://maps.googleapis.com/maps/api/staticmap?"
                     "center={0},{1}&markers=color:red%7C{0},{1}&size=400x400"
                     .format(location.latitude, location.longitude))
        return hangouts_pb2.EmbedItem(
                type=[hangouts_pb2.ITEM_TYPE_PLACE_V2,
                      hangouts_pb2.ITEM_TYPE_PLACE,
                      hangouts_pb2.ITEM_TYPE_THING],
                id="and0",
                place=hangouts_pb2.Place(
                    name=location.name,
                    address=hangouts_pb2.EmbedItem(postal_address=address),
                    geo=hangouts_pb2.EmbedItem(geo_coordinates=geo),
                    representative_image=hangouts_pb2.EmbedItem(
                        type=[hangouts_pb2.ITEM_TYPE_PLACE,
                              hangouts_pb2.ITEM_TYPE_THING],
                        id=image_url,
                        image=hangouts_pb2.EmbedItem.Image(url=image_url)))).place


class HangoutsMessage(immp.Message):
    """
    Message originating from Hangouts.
    """

    @classmethod
    def from_event(cls, hangouts, event):
        """
        Convert a :class:`hangups.ChatMessageEvent` into a :class:`.Message`.

        Args:
            hangouts (.HangoutsPlug):
                Related plug instance that provides the event.
            event (hangups.ChatMessageEvent):
                Hangups message event emitted from a conversation.

        Returns:
            .HangoutsMessage:
                Parsed message object.
        """
        user = HangoutsUser.from_user(hangouts, hangouts._users.get_user(event.user_id))
        action = False
        joined = None
        left = None
        title = None
        attachments = []
        if isinstance(event, hangups.ChatMessageEvent):
            segments = [HangoutsSegment.from_segment(segment) for segment in event.segments]
            text = immp.RichText(segments)
            if any(a.type == 4 for a in event._event.chat_message.annotation):
                # This is a /me message sent from desktop Hangouts.
                action = True
                # The user's first name prefixes the message text, so try to strip that.
                if user.real_name:
                    # We don't have a clear-cut first name, so try to match parts of names.
                    # Try the full name first, then split successive words off the end.
                    parts = user.real_name.split()
                    start = text[0].text
                    for pos in range(len(parts), 0, -1):
                        sub_name = " ".join(parts[:pos])
                        if start.startswith(sub_name):
                            text[0].text = start[len(sub_name) + 1:]
                            break
                    else:
                        # Couldn't match the user's name to the message text.
                        pass
            for attach in event.attachments:
                attachments.append(immp.File(type=immp.File.Type.image, source=attach))
            for attach in event._event.chat_message.message_content.attachment:
                embed = attach.embed_item
                if any(place in embed.type for place in
                       (hangouts_pb2.ITEM_TYPE_PLACE, hangouts_pb2.ITEM_TYPE_PLACE_V2)):
                    location = HangoutsLocation.from_embed(embed)
                    if str(text) == ("https://maps.google.com/maps?q={0},{1}"
                                     .format(location.latitude, location.longitude)):
                        text = None
                    attachments.append(location)
        elif isinstance(event, hangups.MembershipChangeEvent):
            action = True
            is_join = event.type_ == hangouts_pb2.MEMBERSHIP_CHANGE_TYPE_JOIN
            parts = [HangoutsUser.from_user(hangouts, hangouts._users.get_user(part_id))
                     for part_id in event.participant_ids]
            if len(parts) == 1 and parts[0].id == user.id:
                # Membership event is a user acting on themselves.
                segments = [HangoutsSegment("{} the hangout"
                                            .format("joined" if is_join else "left"))]
            else:
                segments = [HangoutsSegment("added " if is_join else "removed ")]
                for part in parts:
                    link = "https://hangouts.google.com/chat/person/{}".format(part.id)
                    segments.append(HangoutsSegment(part.real_name, bold=True, link=link))
                    segments.append(HangoutsSegment(", "))
                # Replace trailing comma.
                segments[-1].text = " {} the hangout".format("to" if is_join else "from")
            if is_join:
                joined = parts
            else:
                left = parts
        elif isinstance(event, hangups.OTREvent):
            action = True
            is_history = (event.new_otr_status ==
                          hangouts_pb2.OFF_THE_RECORD_STATUS_ON_THE_RECORD)
            segments = [HangoutsSegment("{}abled hangout message history"
                                        .format("en" if is_history else "dis"))]
        elif isinstance(event, hangups.RenameEvent):
            action = True
            title = event.new_name
            segments = [HangoutsSegment("renamed the hangout to "),
                        HangoutsSegment(event.new_name, bold=True)]
        elif isinstance(event, hangups.GroupLinkSharingModificationEvent):
            action = True
            is_shared = event.new_status == hangouts_pb2.GROUP_LINK_SHARING_STATUS_ON
            segments = [HangoutsSegment("{}abled joining the hangout by link"
                                        .format("en" if is_shared else "dis"))]
        elif isinstance(event, hangups.HangoutEvent):
            action = True
            texts = {hangouts_pb2.HANGOUT_EVENT_TYPE_START: "started a call",
                     hangouts_pb2.HANGOUT_EVENT_TYPE_END: "ended the call",
                     hangouts_pb2.HANGOUT_EVENT_TYPE_JOIN: "joined the call",
                     hangouts_pb2.HANGOUT_EVENT_TYPE_LEAVE: "left the call"}
            try:
                segments = [HangoutsSegment(texts[event.event_type])]
            except KeyError:
                raise NotImplementedError
        else:
            raise NotImplementedError
        if not isinstance(event, hangups.ChatMessageEvent):
            text = immp.RichText(segments)
        return immp.SentMessage(id=event.id_,
                                channel=immp.Channel(hangouts, event.conversation_id),
                                at=event.timestamp,
                                text=text,
                                user=user,
                                action=action,
                                joined=joined,
                                left=left,
                                title=title,
                                attachments=attachments,
                                raw=event)


class HangoutsPlug(immp.Plug):
    """
    Plug for `Google Hangouts <https://hangouts.google.com>`_.
    """

    network_name = "Hangouts"

    @property
    def network_id(self):
        return "hangouts:{}".format(self._bot_user) if self._bot_user else None

    def __init__(self, name, config, host):
        super().__init__(name, _Schema.config(config), host)
        self._client = None
        self._looped = None
        self._starting = Condition()
        self._closing = False
        self._users = self._convs = None
        self._bot_user = None

    async def _loop(self):
        while True:
            try:
                await self._client.connect()
            except CancelledError:
                log.debug("Cancel request for plug %r loop", self.name)
                return
            except Exception as e:
                log.debug("Unexpected client disconnect: %r", e)
            if self._closing:
                return
            log.debug("Reconnecting in 3 seconds")
            await sleep(3)

    async def _connect(self):
        log.debug("Retrieving users and conversations")
        self._users, self._convs = await hangups.build_user_conversation_list(self._client)
        self._convs.on_event.add_observer(self._event)
        resp = await self._client.get_self_info(hangouts_pb2.GetSelfInfoRequest(
            request_header=self._client.get_request_header()))
        self._bot_user = resp.self_entity.id.chat_id
        async with self._starting:
            self._starting.notify_all()

    async def _event(self, event):
        try:
            sent = HangoutsMessage.from_event(self, event)
        except NotImplementedError:
            log.warn("Skipping unimplemented %r event type", event.__class__.__name__)
        else:
            log.debug("Queueing new message event")
            self.queue(sent)
        if self.config["read"]:
            await self._convs.get(event.conversation_id).update_read_timestamp()

    async def start(self):
        await super().start()
        self._closing = False
        self._client = hangups.Client(hangups.get_auth_stdin(self.config["cookie"]))
        self._client.on_connect.add_observer(self._connect)
        log.debug("Connecting client")
        self._looped = ensure_future(self._loop())
        async with self._starting:
            # Block until users and conversations are loaded.
            await self._starting.wait()
        log.debug("Listening for events")

    async def stop(self):
        await super().stop()
        self._closing = True
        if self._client:
            log.debug("Requesting client disconnect")
            await self._client.disconnect()
            self._client = None
        if self._looped:
            self._looped.cancel()
            self._looped = None
        self._bot_user = None

    async def user_from_id(self, id):
        user = self._users.get_user(hangups.user.UserID(chat_id=id, gaia_id=id))
        if user:
            return HangoutsUser.from_user(self, user)
        request = hangouts_pb2.GetEntityByIdRequest(
            request_header=self._client.get_request_header(),
            batch_lookup_spec=[hangouts_pb2.EntityLookupSpec(gaia_id=id)])
        response = await self._client.get_entity_by_id(request)
        if response.entity:
            return HangoutsUser.from_entity(self, response.entity)
        else:
            return None

    async def user_is_system(self, user):
        return user.id == self._bot_user

    def _filter_channels(self, type):
        convs = self._convs.get_all(include_archived=True)
        return (immp.Channel(self, conv.id_) for conv in convs if conv._conversation.type == type)

    async def public_channels(self):
        return list(self._filter_channels(hangouts_pb2.CONVERSATION_TYPE_GROUP))

    async def private_channels(self):
        return list(self._filter_channels(hangouts_pb2.CONVERSATION_TYPE_ONE_TO_ONE))

    async def channel_for_user(self, user):
        for channel in self._filter_channels(hangouts_pb2.CONVERSATION_TYPE_ONE_TO_ONE):
            if any(part.id == user.id for part in await channel.members()):
                return channel
        request = hangouts_pb2.CreateConversationRequest(
            request_header=self._client.get_request_header(),
            type=hangouts_pb2.CONVERSATION_TYPE_ONE_TO_ONE,
            client_generated_id=self._client.get_client_generated_id(),
            invitee_id=[hangouts_pb2.InviteeID(gaia_id=user.id)])
        response = await self._client.create_conversation(request)
        return immp.Channel(self, response.conversation.conversation_id.id)

    async def channel_is_private(self, channel):
        try:
            conv = self._convs.get(channel.source)
        except KeyError:
            return False
        else:
            return conv._conversation.type == hangouts_pb2.CONVERSATION_TYPE_ONE_TO_ONE

    async def channel_title(self, channel):
        try:
            return self._convs.get(channel.source).name
        except KeyError:
            return None

    async def channel_link(self, channel):
        return "https://hangouts.google.com/chat/{}".format(channel.source)

    async def channel_rename(self, channel, title):
        try:
            conv = self._convs.get(channel.source)
        except KeyError:
            return None
        else:
            if not conv.name == title:
                await conv.rename(title)

    async def channel_members(self, channel):
        try:
            conv = self._convs.get(channel.source)
        except KeyError:
            return None
        else:
            return [HangoutsUser.from_user(self, user) for user in conv.users]

    async def channel_invite(self, channel, user):
        try:
            conv = self._convs.get(channel.source)
        except KeyError:
            return
        request = hangouts_pb2.AddUserRequest(
            request_header=self._client.get_request_header(),
            event_request_header=conv._get_event_request_header(),
            invitee_id=[hangouts_pb2.InviteeID(gaia_id=user.id)])
        await self._client.add_user(request)

    async def channel_remove(self, channel, user):
        try:
            conv = self._convs.get(channel.source)
        except KeyError:
            return
        request = hangouts_pb2.RemoveUserRequest(
            request_header=self._client.get_request_header(),
            event_request_header=conv._get_event_request_header(),
            participant_id=hangouts_pb2.ParticipantId(gaia_id=user.id))
        await self._client.remove_user(request)

    async def _next_batch(self, conv, before_id):
        # Conversation.get_events() should, if the target is the oldest message in the current
        # batch, fetch the next whole batch and return that, or else return everything before the
        # target.  However, at the end of the message history, it sometimes returns an arbitrary
        # batch instead.  Return fetched messages from Conversation.events directly instead.
        ids = [event.id_ for event in conv.events]
        if before_id not in ids:
            return None
        if ids[0] == before_id:
            # Target is the oldest message cached, so there may be more -- try for another batch.
            await conv.get_events(before_id)
            ids = [event.id_ for event in conv.events]
        # Take all events older than the target.
        events = conv.events[:ids.index(before_id)]
        return [HangoutsMessage.from_event(self, event) for event in events]

    async def channel_history(self, channel, before=None):
        try:
            conv = self._convs.get(channel.source)
        except KeyError:
            return []
        if not conv.events:
            return []
        if not before:
            if len(conv.events) == 1:
                # Only the initial message cached, try to fetch a first batch.
                await conv.get_events(conv.events[0].id_)
            # Return all cached events.
            return [HangoutsMessage.from_event(self, event) for event in conv.events]
        ids = [event.id_ for event in conv.events]
        if before.id in ids:
            return await self._next_batch(conv, before.id)
        # Hangouts has no way to query for an event by ID, only by timestamp.  Instead, we'll try a
        # few times to retrieve it further down the message history.
        for i in range(10):
            log.debug("Fetching batch %i of events to find %r", i + 1, before.id)
            events = await conv.get_events(conv.events[0].id_)
            ids = [event.id_ for event in events]
            if not ids:
                # No further messages, we've hit the end of the message history.
                return []
            elif before.id in ids:
                return await self._next_batch(conv, before.id)
        # Maxed out on attempts but didn't find the requested message.
        return []

    async def _get_event(self, conv, event_id):
        ids = [event.id_ for event in conv.events]
        try:
            return conv.get_event(event_id)
        except KeyError:
            pass
        # Hangouts has no way to query for an event by ID, only by timestamp.  Instead, we'll try a
        # few times to retrieve it further down the message history.
        for i in range(10):
            log.debug("Fetching batch %i of events to find %r", i + 1, event_id)
            events = await conv.get_events(conv.events[0].id_)
            ids = [event.id_ for event in events]
            if not ids:
                # No further messages, we've hit the end of the message history.
                return []
            elif event_id in ids:
                return events[ids.index(event_id)]
        # Maxed out on attempts but didn't find the requested message.
        return None

    async def _resolve_msg(self, conv, msg):
        if isinstance(msg, immp.Message):
            return msg
        elif isinstance(msg, immp.Receipt):
            # We have the message reference but not the content.
            event = await self._get_event(conv, msg.id)
            if not event:
                return None
            sent = HangoutsMessage.from_event(self, event)
            # As we only use this for rendering the message again, we shouldn't add a second
            # layer of authorship if we originally sent the message being retrieved.
            if sent.user.id == self._bot_user:
                sent.user = None
            return sent
        else:
            return None

    async def _upload(self, attach):
        async with (await attach.get_content()) as img_content:
            # Hangups expects a file-like object with a synchronous read() method.
            # NB. The whole file is read into memory by Hangups anyway.
            # Filename must be present, else Hangups will try (and fail) to read the path.
            photo = await self._client.upload_image(BytesIO(await img_content.read()),
                                                    filename=attach.title or "image.png")
        return hangouts_pb2.ExistingMedia(photo=hangouts_pb2.Photo(photo_id=photo))

    @classmethod
    def _serialise(cls, segments):
        output = []
        for segment in segments:
            output += HangoutsSegment.to_segments(segment)
        return [segment.serialize() for segment in output]

    def _request(self, conv, segments=None, media=None, place=None):
        return hangouts_pb2.SendChatMessageRequest(
            request_header=self._client.get_request_header(),
            event_request_header=conv._get_event_request_header(),
            message_content=hangouts_pb2.MessageContent(segment=segments) if segments else None,
            existing_media=media,
            location=hangouts_pb2.Location(place=place) if place else None)

    async def _requests(self, conv, msg):
        uploads = []
        images = []
        places = []
        for attach in msg.attachments:
            if isinstance(attach, immp.File) and attach.type == immp.File.Type.image:
                uploads.append(self._upload(attach))
            elif isinstance(attach, immp.Location):
                places.append(HangoutsLocation.to_place(attach))
        if uploads:
            images = await gather(*uploads)
        requests = []
        if msg.text or msg.reply_to:
            edited = msg.edited if isinstance(msg, immp.Receipt) else False
            segments = self._serialise(msg.render(link_name=False, edit=edited, quote_reply=True))
            media = None
            if len(images) == 1:
                # Attach the only image to the message text.
                media = images.pop()
            requests.append(self._request(conv, segments, media))
        if images:
            segments = []
            if msg.user:
                label = immp.Message(user=msg.user, text="sent an image", action=True)
                segments = self._serialise(label.render(link_name=False))
            # Send any additional media items in their own separate messages.
            for media in images:
                requests.append(self._request(conv, segments, media))
        if places:
            # Send each location separately.
            for place in places:
                requests.append(self._request(conv, place=place))
            # Include a label only if we haven't sent a text message earlier.
            if msg.user and not msg.text:
                label = immp.Message(user=msg.user, text="sent a location", action=True)
                segments = self._serialise(label.render(link_name=False))
                requests.append(self._request(conv, segments))
        return requests

    async def put(self, channel, msg):
        conv = self._convs.get(channel.source)
        # Attempt to find sources for referenced messages.
        clone = copy(msg)
        clone.reply_to = await self._resolve_msg(conv, clone.reply_to)
        requests = []
        for attach in clone.attachments:
            # Generate requests for attached messages first.
            if isinstance(attach, immp.Message):
                requests += await self._requests(conv, await self._resolve_msg(conv, attach))
        own_requests = await self._requests(conv, clone)
        if requests and not own_requests:
            # Forwarding a message but no content to show who forwarded it.
            info = immp.Message(user=clone.user, action=True, text="forwarded a message")
            own_requests = await self._requests(conv, info)
        requests += own_requests
        events = []
        for request in requests:
            events.append(await self._client.send_chat_message(request))
        return [event.created_event.event_id for event in events]



from asyncio import Condition
from enum import Enum
from functools import reduce
from importlib import import_module
import re
import time

from .error import ConfigError


class IDGen:
    """
    Generator of generic timestamp-based identifiers.

    IDs are guaranteed unique for the lifetime of the application -- two successive calls will
    yield two different identifiers.
    """

    def __init__(self):
        self.last = 0

    def __call__(self):
        """
        Make a new identifier.

        Returns:
            str:
                Newly generated identifier.
        """
        new = max(self.last + 1, int(time.time()))
        self.last = new
        return str(new)

    def __repr__(self):
        return "<{}: {} -> {}>".format(self.__class__.__name__, self.last, self())


def resolve_import(path):
    """
    Take the qualified name of a Python class, and return the physical class object.

    Args:
        path (str):
            Dotted Python class name, e.g. ``<module path>.<class name>``.

    Returns:
        type:
            Class object imported from module.
    """
    module, class_ = path.rsplit(".", 1)
    return getattr(import_module(module), class_)


def pretty_str(cls):
    """
    Class decorator to provide a default :meth:`__str__` based on the contents of :attr:`__dict__`.
    """

    def nest_str(obj):
        if isinstance(obj, dict):
            return "{{...{}...}}".format(len(obj)) if obj else "{}"
        elif isinstance(obj, list):
            return "[...{}...]".format(len(obj)) if obj else "[]"
        else:
            return str(obj)

    def __str__(self):
        args = "\n".join("{}: {}".format(k, nest_str(v).replace("\n", "\n" + " " * (len(k) + 2)))
                         for k, v in self.__dict__.items() if not k.startswith("_"))
        return "[{}]\n{}".format(self.__class__.__name__, args)

    cls.__str__ = __str__
    return cls


def _no_escape(char):
    # Fail a match if the next character is escaped by a backslash.
    return r"(?<!\\)(?:\\\\)*{}".format(char)


def escape(raw, *chars):
    """
    Prefix special characters with backslashes, suitable for encoding in a larger string
    delimiting those characters.

    Args:
        raw (str):
            Unsafe input string.
        chars (str list):
            Control characters to be escaped.

    Returns:
        str:
            Escaped input string.
    """
    args = (raw, r"\\", *chars)
    return reduce(lambda current, char: current.replace(char, r"\{}".format(char)), args)


def unescape(raw, *chars):
    """
    Inverse of :func:`escape`, remove backslashes escaping special characters.

    Args:
        raw (str):
            Escaped input string.
        chars (str list):
            Control characters to be unescaped.

    Returns:
        str:
            Raw unescaped string.
    """
    args = (raw, *chars, r"\\")
    return reduce(lambda current, char: re.sub(_no_escape(r"\\{}".format(char)),
                                               char, current), args)


class ConfigProperty:
    """
    Data descriptor to present config from :class:`.Openable` instances using the actual objects
    stored in a :class:`.Host`.
    """

    def __init__(self, cls=None, key=None):
        self._cls = cls
        self._key = key

    def __set_name__(self, owner, name):
        if not self._key:
            self._key = name.lstrip("_")

    @classmethod
    def _describe(cls, spec):
        if isinstance(spec, list):
            return "[{}]".format(", ".join(cls._describe(inner) for inner in spec))
        elif isinstance(spec, tuple):
            return "{{{}}}".format(", ".join(inner.__name__ for inner in spec))
        else:
            return spec.__name__

    def _from_host(self, instance, name, cls=None):
        try:
            obj = instance.host[name]
        except KeyError:
            raise ConfigError("No object {} on host".format(repr(name))) from None
        if cls and not isinstance(obj, cls):
            raise ConfigError("Reference {} not instance of {}"
                              .format(repr(name), self._describe(cls)))
        else:
            return obj

    def __get__(self, instance, owner):
        if not instance:
            return self
        value = instance.config.get(self._key)
        if not value:
            return None
        if isinstance(self._cls, list):
            obj = tuple(self._from_host(instance, name, tuple(self._cls)) for name in value)
        else:
            obj = self._from_host(instance, value, self._cls)
        return obj

    def __repr__(self):
        return "<{}: {}{}>".format(self.__class__.__name__, repr(self._key),
                                   " {}".format(self._describe(self._cls)) if self._cls else "")


class OpenState(Enum):
    """
    Readiness status for instances of :class:`Openable`.
    """
    inactive = 0
    starting = 1
    active = 2
    stopping = 3


class Openable:
    """
    Abstract class to provide open and close hooks.  Subclasses should implement :meth:`start`
    and :meth:`stop`, whilst users should make use of :meth:`open` and :meth:`close`.

    Attributes:
        state (.OpenState):
            Current status of this resource.
    """

    def __init__(self):
        self.state = OpenState.inactive
        self._changing = Condition()

    async def open(self):
        """
        Open this resource ready for use.  Does nothing if already open, but raises
        :class:`RuntimeError` if currently changing state.
        """
        if self.state == OpenState.active:
            return
        elif self.state != OpenState.inactive:
            raise RuntimeError("Can't open when already closing")
        self.state = OpenState.starting
        await self.start()
        self.state = OpenState.active

    async def start(self):
        """
        Perform any operations needed to open this resource.
        """

    async def close(self):
        """
        Close this resource after it's used.  Does nothing if already closed, but raises
        :class:`RuntimeError` if currently changing state.
        """
        if self.state == OpenState.inactive:
            return
        elif self.state != OpenState.active:
            raise RuntimeError("Can't close when already opening/closing")
        self.state = OpenState.stopping
        await self.stop()
        self.state = OpenState.inactive

    async def stop(self):
        """
        Perform any operations needed to close this resource.
        """



"""
Backbone for other hooks to process commands contained in channel messages.

Config:
    prefix (str):
        Characters at the start of a message to denote commands.  Use a single character to
        make commands top-level (e.g. ``"?"`` would allow commands like ``?help``), or a string
        followed by a space for subcommands (e.g. ``"!bot "`` for ``!bot help``).
    return-errors (bool):
        ``True`` to send unhandled exceptions raised by commands back to the source channel
        (``False`` by default).
    sets ((str, str list) dict):
        Subsets of hook commands by name, to restrict certain features.
    mapping (str, dict) dict):
        Named config groups to enable commands in selected channels.

        groups (str list):
            List of groups (plugs and channels) to process public commands in.
        hooks (str list):
            List of hooks to enable commands for.
        sets (str list):
            List of command sets to enable.
        admins ((str, str list) dict):
            Users authorised to execute administrative commands, a mapping of network identifiers
            to lists of user identifiers.

The binding works by making commands exposed by all listed hooks available to all listed channels,
and to the private channels of all listed plugs.  Note that the channels need not belong to any
hook-specific config -- you can, for example, bind some commands to an admin-only channel
elsewhere.  Multiple groups can be used for fine-grained control.
"""

from collections import defaultdict
from copy import copy
from enum import Enum
import inspect
import logging
import re
import shlex

from voluptuous import ALLOW_EXTRA, Any, Optional, Schema

import immp


log = logging.getLogger(__name__)


class _Schema:

    def _key(name, default=dict):
        return Optional(name, default=default)

    config = Schema({"prefix": [str],
                     _key("return-errors", False): bool,
                     _key("sets"): Any({}, {str: {str: [str]}}),
                     "mapping": {str: {_key("groups", list): [str],
                                       _key("hooks", list): [str],
                                       _key("sets", list): [str],
                                       _key("admins"): Any({}, {str: [str]})}}},
                    extra=ALLOW_EXTRA, required=True)


class BadUsage(immp.HookError):
    """
    May be raised from within a command to indicate that the arguments were invalid.
    """


class CommandParser(Enum):
    """
    Constants representing the method used to parse the argument text following a used command.

    Attributes:
        spaces:
            Split using :meth:`str.split`, for simple inputs breaking on whitespace characters.
        shlex:
            Split using :func:`shlex.split`, which allows quoting of multi-word arguments.
        none:
            Don't split the argument, just provide the rich message text as-is.
        hybrid:
            Split as with ``spaces`` up to the number of accepted arguments, and return rich text
            in the last argument.  If optional arguments are present, only the last will receive
            the rich text, or none of them if not filled in.
    """
    spaces = 0
    shlex = 1
    none = 2
    hybrid = 3


class CommandScope(Enum):
    """
    Constants representing the types of conversations a command is available in.

    Attributes:
        anywhere:
            All configured channels.
        private:
            Only private channels, as configured per-plug.
        shared:
            Only non-private channels, as configured per-channel.
    """
    anywhere = 0
    private = 1
    shared = 2


class CommandRole(Enum):
    """
    Constants representing the types of users a command is available in.

    Attributes:
        anyone:
            All configured channels.
        admin:
            Only authorised users in the command group.
    """
    anyone = 0
    admin = 1


@immp.pretty_str
class BoundCommand:
    """
    Wrapper object returned when accessing a command via a :class:`.Hook` instance, similar to
    :class:`types.MethodType`.

    This object is callable, which invokes the command's underlying method against the bound hook.
    """

    def __init__(self, hook, cmd):
        self.hook = hook
        self.cmd = cmd

    def applicable(self, channel, user, private, admin):
        """
        Test the availability of the current command based on the scope and role.

        Args:
            channel (.Channel):
                Source channel where the command will be executed.
            user (.User):
                Author of the message to trigger the command.
            private (bool):
                Result of :meth:`.Channel.is_private`.
            admin (bool):
                ``True`` if the author is defined as an admin of this :class:`.CommandHook`.

        Returns:
            bool:
                ``True`` if the command may be used.
        """
        if self.scope == CommandScope.private and not private:
            return False
        elif self.scope == CommandScope.shared and private:
            return False
        elif self.role == CommandRole.admin and not admin:
            return False
        elif self.cmd.test:
            return self.cmd.test(self.hook, channel, user)
        else:
            return True

    def complete(self, name, *args):
        """
        Fully-qualify an underlying base command.

        Args:
            name (str):
                Command name, used to access the command when directly following the prefix.
            fixed_args (tuple):
                Additional arguments to pass to the underlying method.

        Returns:
            .BoundCommand:
                Wrapper around the now-fully-qualified instance of the base command.
        """
        return BoundCommand(self.hook, self.cmd.complete(name, *args))

    async def __call__(self, msg, *args):
        if isinstance(self.cmd, BaseCommand) and not isinstance(self.cmd, FullCommand):
            raise ValueError("Command is dynamic")
        else:
            return await self.cmd.fn(self.hook, *self.cmd.fixed_args, msg, *args)

    def __getattr__(self, name):
        # Propagate other attribute access to the unbound command object.
        return getattr(self.cmd, name)

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and
                (self.hook, self.cmd) == (other.hook, other.cmd))

    def __hash__(self):
        return hash((self.hook, self.cmd))

    def __repr__(self):
        return "<{}: {} {}>".format(self.__class__.__name__, repr(self.hook), repr(self.cmd))


@immp.pretty_str
class BaseCommand:
    """
    Container of a command function.  Use the :meth:`command` decorator to wrap a class method and
    convert it into an instance of this class.  Unless dynamic, command objects will be an instance
    of the :class:`.FullCommand` subclass.

    Accessing an instance of this class via the attribute of a containing class will create a new
    :class:`BoundCommand` allowing invocation of the method.

    Attributes:
        fn (method):
            Callback method from a hook to process the command usage.
        parser (.CommandParser):
            Parse mode for the command arguments.
        scope (.CommandScope):
            Accessibility of this command for the different channel types.
        role (.CommandRole):
            Accessibility of this command for different users.
        test (method):
            Additional predicate that can enable or disable a command based on hook state.
        sync_aware (bool):
            ``True`` if the hook is aware of synced channels provided by a :class:`.SyncPlug`.
        fixed (int):
            Number of fixed arguments the underlying method requires.
        doc (str):
            Help for the command, taken from the callback method's docstring if present.
        spec (str):
            Human-readable summary of the arguments required by the method.
    """

    def __init__(self, fn, parser=CommandParser.spaces, scope=CommandScope.anywhere,
                 role=CommandRole.anyone, test=None, sync_aware=False):
        self.fn = fn
        self.parser = parser
        self.scope = scope
        self.role = role
        self.test = test
        self.sync_aware = sync_aware
        # Only positional arguments are produced by splitting the input, there are no keywords.
        if any(param.kind in (inspect.Parameter.KEYWORD_ONLY,
                              inspect.Parameter.VAR_KEYWORD) for param in self._args[1]):
            raise ValueError("Keyword-only command parameters are not supported: {}".format(fn))

    @property
    def _args(self):
        # Skip `self` argument.
        params = tuple(inspect.signature(self.fn).parameters.values())[1:]
        # Split on `msg` argument into fixed and called arguments.
        for i, param in enumerate(params):
            if param.name == "msg":
                return (params[:i], params[i + 1:])
        else:
            raise ValueError("Command method doesn't accept a `msg` parameter")

    @property
    def fixed(self):
        return len(self._args[0])

    @property
    def min(self):
        return len([arg for arg in self._args[1] if arg.default is inspect.Parameter.empty])

    @property
    def max(self):
        return len(self._args[1])

    @property
    def doc(self):
        return inspect.cleandoc(self.fn.__doc__) if self.fn.__doc__ else None

    @property
    def spec(self):
        parts = []
        for param in self._args[1]:
            if param.kind in (inspect.Parameter.POSITIONAL_ONLY,
                              inspect.Parameter.POSITIONAL_OR_KEYWORD):
                parts.append(("<{}>" if param.default is inspect.Parameter.empty else "[{}]")
                             .format(param.name))
            elif param.kind == inspect.Parameter.VAR_POSITIONAL:
                parts.append("[{}...]".format(param.name))
        return " ".join(parts)

    def parse(self, args):
        """
        Convert a string of multiple arguments into a list according to the chosen parse mode.

        Args:
            args (.RichText):
                Trailing argument text from a message.

        Returns:
            (str or RichText) list:
                Parsed arguments.
        """
        if not args:
            return []
        if self.parser == CommandParser.spaces:
            return str(args).split()
        elif self.parser == CommandParser.shlex:
            return shlex.split(str(args))
        elif self.parser == CommandParser.hybrid:
            filled = self.max
            parts = str(args).split(maxsplit=filled - 1)
            if len(parts) < filled:
                return parts
            full = re.split(r"(\s+)", str(args))
            index = filled * 2
            if full[0]:
                index -= 2
            offset = len("".join(full[:index]))
            return parts[:-1] + [args[offset::True]]
        else:
            return [args]

    def valid(self, *args):
        """
        Test the validity of the given arguments against the command's underlying method.  Raises
        :class:`ValueError` if the arguments don't match the signature.

        Args:
            args (str list):
                Parsed arguments.
        """
        params = self._args[1]
        varargs = len([arg for arg in params if arg.kind is inspect.Parameter.VAR_POSITIONAL])
        required = self.min - varargs
        if len(args) < required:
            raise ValueError("Expected at least {} args, got {}".format(required, len(args)))
        if len(args) > len(params) and not varargs:
            raise ValueError("Expected at most {} args, got {}".format(len(params), len(args)))

    def complete(self, name, *args):
        """
        Fully-qualify a base command.

        Args:
            name (str):
                Command name, used to access the command when directly following the prefix.
            fixed_args (tuple):
                Additional arguments to pass to the underlying method.

        Returns:
            .FullCommand:
                Fully-qualified instance of this base command.
        """
        return FullCommand(name, self.fn, self.parser, self.scope, self.role, self.test,
                           self.sync_aware, args)

    def __get__(self, instance, owner):
        return BoundCommand(instance, self) if instance else self

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.fn == other.fn

    def __hash__(self):
        return hash(self.fn)

    def __repr__(self):
        return "<{}: {}, {} {}>".format(self.__class__.__name__, self.scope.name,
                                        self.role.name, self.fn)


class FullCommand(BaseCommand):
    """
    Fully-qualified and named command, an extension of :class:`.BaseCommand` to make it callable.

    Attributes:
        name (str):
            Command name, used to access the command when directly following the prefix.
        fixed_args (tuple):
            Additional arguments to pass to the underlying method.
    """

    def __init__(self, name, fn, parser=CommandParser.spaces, scope=CommandScope.anywhere,
                 role=CommandRole.anyone, test=None, sync_aware=False, fixed_args=()):
        super().__init__(fn, parser, scope, role, test, sync_aware)
        self.name = name.lower()
        self.fixed_args = fixed_args
        if len(self.fixed_args) != self.fixed:
            raise ValueError("Expecting {} fixed arguments, got {}"
                             .format(self.fixed, len(self.fixed_args)))

    def complete(self, name, *args):
        raise ValueError("Already a full command")

    def __eq__(self, other):
        return super().__eq__(other) and self.name == other.name

    def __hash__(self):
        return hash((super().__hash__(), self.name))

    def __repr__(self):
        return "<{}: {} @ {}, {} {}>".format(self.__class__.__name__, self.name, self.scope.name,
                                             self.role.name, self.fn)


def command(name=None, parser=CommandParser.spaces, scope=CommandScope.anywhere,
            role=CommandRole.anyone, test=None, sync_aware=False):
    """
    Decorator: mark up a hook method as a command.

    This doesn't return the original function, rather a :class:`.BaseCommand` object.

    The method's accepted arguments must start with ``self``, and must accept one called ``msg``,
    which will be set to the :class:`.Message` instance that triggered this command.

    A *dynamic* command is an unnamed instance that can be used to create multiple commands with
    the same underlying method call.  These commands are not collected automatically -- the parent
    hook must implement :class:`.DynamicCommands`.

    Arguments:
        name (str):
            Command name, used to access the command when directly following the prefix.
        parser (.CommandParser):
            Parse mode for the command arguments.
        scope (.CommandScope):
            Accessibility of this command for the different channel types.
        role (.CommandRole):
            Accessibility of this command for different users.
        test (method):
            Additional predicate that can enable or disable a command based on hook state.
        sync_aware (bool):
            ``True`` if the hook is aware of synced channels provided by a :class:`.SyncPlug`.  In
            this case, the command handler will receive the native channel rather than the virtual
            sync channel.  See :meth:`.SyncPlug.sync_for` for resolving this to a virtual channel.
    """
    def wrap(fn):
        args = (fn, parser, scope, role, test, sync_aware)
        return FullCommand(name, *args) if name else BaseCommand(*args)
    return wrap


class DynamicCommands:
    """
    Interface for commands generated at runtime.  Typically combined with additional arguments in a
    command method's signature.
    """

    def commands(self):
        """
        Provide additional commands dependent on state or config.  Dynamic commands must be filled
        in using :meth:`.BaseCommand.complete` to provide their name and any fixed arguments.

        Returns:
            .BoundCommand set:
                Set of available commands.
        """
        raise NotImplementedError


class CommandHook(immp.Hook):
    """
    Generic command handler for other hooks.
    """

    def __init__(self, name, config, host):
        super().__init__(name, _Schema.config(config), host)
        # Avoiding circular dependency between commands and sync -- use the full path to populate
        # that attribute path in the global `immp` import for later (so unused here).
        import immp.hook.sync

    def discover(self, hook):
        """
        Inspect a :class:`.Hook` instance, scanning its attributes for commands.

        Returns:
            (str, .BoundCommand) dict:
                Commands provided by this hook, keyed by name.
        """
        if hook.state != immp.OpenState.active:
            return {}
        attrs = [getattr(hook, attr) for attr in dir(hook)]
        cmds = {cmd.name: cmd for cmd in attrs if isinstance(cmd, BoundCommand)
                and isinstance(cmd.cmd, FullCommand)}
        if isinstance(hook, DynamicCommands):
            cmds.update({cmd.name: cmd for cmd in hook.commands()})
        return cmds

    def _mapping_cmds(self, mapping, channel, user, private):
        cmdgroup = set()
        admin = user.plug and user.id in (mapping["admins"].get(user.plug.name) or [])
        for name in mapping["hooks"]:
            cmdgroup.update(set(self.discover(self.host.hooks[name]).values()))
        for label in mapping["sets"]:
            for name, cmdset in self.config["sets"][label].items():
                discovered = self.discover(self.host.hooks[name])
                cmdgroup.update(set(discovered[cmd] for cmd in cmdset))
        return {cmd for cmd in cmdgroup if cmd.applicable(channel, user, private, admin)}

    async def commands(self, channel, user):
        """
        Retrieve all commands, and filter against the mappings.

        Args:
            channel (.Channel):
                Source channel where the command will be executed.
            user (.User):
                Author of the message to trigger the command.

        Returns:
            (str, .BoundCommand) dict:
                Commands provided by all hooks, in this channel for this user, keyed by name.
        """
        log.debug("Collecting commands for %r in %r", channel, user)
        if isinstance(channel, immp.Plug):
            # Look for commands for a generic channel.
            plug = channel
            channel = immp.Channel(plug, "")
            private = False
        else:
            plug = None
            private = await channel.is_private()
        mappings = []
        for mapping in self.config["mapping"].values():
            for label in mapping["groups"]:
                group = self.host.groups[label]
                if plug and group.has_plug(plug, "anywhere", "named"):
                    mappings.append(mapping)
                elif not plug and await group.has_channel(channel):
                    mappings.append(mapping)
        cmds = set()
        for mapping in mappings:
            cmds.update(self._mapping_cmds(mapping, channel, user, private))
        mapped = {cmd.name: cmd for cmd in cmds}
        if len(cmds) > len(mapped):
            # Mapping by name silently overwrote at least one command with a duplicate name.
            raise immp.ConfigError("Multiple applicable commands with the same name")
        return mapped

    @command("help", sync_aware=True)
    async def help(self, msg, command=None):
        """
        List all available commands in this channel, or show help about a single command.
        """
        if await msg.channel.is_private():
            current = None
            private = msg.channel
        else:
            current = msg.channel
            private = await msg.user.private_channel()
        parts = defaultdict(dict)
        if current:
            parts[current] = await self.commands(current, msg.user)
        if private:
            parts[private] = await self.commands(private, msg.user)
            for name in parts[private]:
                parts[current].pop(name, None)
        parts[None] = await self.commands(msg.channel.plug, msg.user)
        for name in parts[None]:
            if private:
                parts[private].pop(name, None)
            if current:
                parts[current].pop(name, None)
        full = dict(parts[None])
        full.update(parts[current])
        full.update(parts[private])
        if command:
            try:
                cmd = full[command]
            except KeyError:
                text = "\N{CROSS MARK} No such command"
            else:
                text = immp.RichText([immp.Segment(cmd.name, bold=True)])
                if cmd.spec:
                    text.append(immp.Segment(" {}".format(cmd.spec)))
                if cmd.doc:
                    text.append(immp.Segment(":", bold=True),
                                immp.Segment("\n"),
                                *immp.RichText.unraw(cmd.doc, self.host))
        else:
            titles = {None: [immp.Segment("Global commands", bold=True)]}
            if private:
                titles[private] = [immp.Segment("Private commands", bold=True)]
            if current:
                titles[current] = [immp.Segment("Commands for ", bold=True),
                                   immp.Segment(await current.title(), bold=True, italic=True)]
            text = immp.RichText()
            for channel, cmds in parts.items():
                if not cmds:
                    continue
                if text:
                    text.append(immp.Segment("\n"))
                text.append(*titles[channel])
                for name, cmd in sorted(cmds.items()):
                    text.append(immp.Segment("\n- {}".format(name)))
                    if cmd.spec:
                        text.append(immp.Segment(" {}".format(cmd.spec), italic=True))
        await msg.channel.send(immp.Message(text=text))

    async def on_receive(self, sent, source, primary):
        await super().on_receive(sent, source, primary)
        if not primary or not sent.user or not sent.text or sent != source:
            return
        plain = str(sent.text)
        raw = None
        for prefix in self.config["prefix"]:
            if plain.lower().startswith(prefix):
                raw = plain[len(prefix):].split(maxsplit=1)
                break
        if not raw:
            return
        # Sync integration: exclude native channels of syncs from command execution.
        if isinstance(sent.channel.plug, immp.hook.sync.SyncPlug):
            log.debug("Suppressing command in virtual sync channel: %r", sent.channel)
            return
        synced = immp.hook.sync.SyncPlug.any_sync(self.host, sent.channel)
        if synced:
            log.debug("Mapping command channel: %r -> %r", sent.channel, synced)
        name = raw[0].lower()
        trailing = sent.text[-len(raw[1])::True] if len(raw) == 2 else None
        cmds = await self.commands(sent.channel, sent.user)
        try:
            cmd = cmds[name]
        except KeyError:
            log.debug("No matches for command name %r in %r", name, sent.channel)
            return
        else:
            log.debug("Matched command in %r: %r", sent.channel, cmd)
        try:
            args = cmd.parse(trailing)
            cmd.valid(*args)
        except ValueError:
            # Invalid number of arguments passed, return the command usage.
            await self.help(sent, name)
            return
        if synced and not cmd.sync_aware:
            msg = copy(sent)
            msg.channel = synced
        else:
            msg = sent
        try:
            log.debug("Executing command: %r %r", sent.channel, sent.text)
            await cmd(msg, *args)
        except BadUsage:
            await self.help(sent, name)
        except Exception as e:
            log.exception("Exception whilst running command: %r", sent.text)
            if self.config["return-errors"]:
                text = ": ".join(filter(None, (e.__class__.__name__, str(e))))
                await sent.channel.send(immp.Message(text="\N{WARNING SIGN} {}".format(text)))



from .core.channel import Channel, Group
from .core.error import ConfigError, HookError, PlugError
from .core.host import Host
from .core.message import File, Location, Message, Receipt, RichText, Segment, SentMessage, User
from .core.hook import Hook, ResourceHook
from .core.plug import Plug
from .core.stream import PlugStream
from .core.util import (escape, pretty_str, resolve_import, unescape, ConfigProperty, IDGen,
                        OpenState, Openable)



"""
Basic identity management for users in different networks.

Config:
    instance (int):
        Unique instance code.
    plugs (str list):
        List of plug names to accept identities for.
    multiple (bool):
        ``True`` (default) to allow linking multiple accounts from the same network.

Commands:
    id-add <name> <pwd>:
        Create a new identity, or link to an existing one from a second user.
    id-rename <name>:
        Rename the current identity.
    id-password <pwd>:
        Update the password for the current identity.
    id-reset:
        Delete the current identity and all linked users.
    id-role <name> [role]:
        List roles assigned to an identity, or add/remove a given role.

This is a local implementation of the identity protocol, providing self-serve identity linking to
users where a user management backend doesn't otherwise exist.

In order to support multiple copies of this hook with overlapping plugs (e.g. including a private
network in some groups), each hook has an instance code.  If a code isn't defined in the config, a
new one will be assigned at startup.  If multiple hooks are in use, it's important to define these
yourself, so that identities remained assigned to the correct instance.

.. note::
    This hook requires an active :class:`.DatabaseHook` to store data.
"""

from asyncio import gather
from hashlib import sha256
import logging

from peewee import CharField, ForeignKeyField, IntegerField
from voluptuous import ALLOW_EXTRA, Any, Optional, Schema

import immp
from immp.hook.access import AccessPredicate
from immp.hook.command import CommandRole, CommandScope, command
from immp.hook.database import BaseModel, DatabaseHook
from immp.hook.identity import Identity, IdentityProvider


CROSS = "\N{CROSS MARK}"
TICK = "\N{WHITE HEAVY CHECK MARK}"


log = logging.getLogger(__name__)


class _Schema:

    config = Schema({Optional("instance", default=None): Any(int, None),
                     "plugs": [str],
                     Optional("multiple", default=True): bool},
                    extra=ALLOW_EXTRA, required=True)


class IdentityGroup(BaseModel):
    """
    Representation of a single identity.

    Attributes:
        instance (int):
            :class:`IdentityHook` instance code.
        name (str):
            Unique display name.
        pwd (str):
            Hashed password, used by the user to authenticate when linking identities.
        links (.IdentityLink iterable):
            All links contained by this group.
    """

    instance = IntegerField()
    name = CharField()
    pwd = CharField()

    class Meta:
        # Uniqueness constraint for each name in each identity instance.
        indexes = ((("instance", "name"), True),)

    @classmethod
    def hash(cls, pwd):
        return sha256(pwd.encode("utf-8")).hexdigest()

    @classmethod
    def select_links(cls):
        return cls.select(cls, IdentityLink).join(IdentityLink)

    async def to_identity(self, host):
        tasks = []
        plugs = {plug.network_id: plug for plug in host.plugs.values()}
        for link in self.links:
            tasks.append(plugs[link.network].user_from_id(link.user))
        users = await gather(*tasks)
        roles = [role.role for role in self.roles]
        return Identity(self.name, users, roles)

    def __repr__(self):
        return "<{}: #{} {}>".format(self.__class__.__name__, self.id, repr(self.name))


class IdentityLink(BaseModel):
    """
    Single link between an identity and a user.

    Attributes:
        group (.IdentityGroup):
            Containing group instance.
        network (str):
            Network identifier that the user belongs to.
        user (str):
            User identifier as given by the plug.
    """

    group = ForeignKeyField(IdentityGroup, related_name="links", on_delete="cascade")
    network = CharField()
    user = CharField()

    def __repr__(self):
        return "<{}: #{} {} @ {} {}>".format(self.__class__.__name__, self.id, repr(self.user),
                                             repr(self.network), repr(self.group))


class IdentityRole(BaseModel):
    """
    Assignment of a role to an identity.

    Attributes:
        group (.IdentityGroup):
            Containing group instance.
        role (str):
            Plain role identifier.
    """

    group = ForeignKeyField(IdentityGroup, related_name="roles", on_delete="cascade")
    role = CharField()

    def __repr__(self):
        return "<{}: #{} {} {}>".format(self.__class__.__name__, self.id, repr(self.role),
                                        repr(self.group))


class LocalIdentityHook(immp.Hook, AccessPredicate, IdentityProvider):
    """
    Hook for managing physical users with multiple logical links across different plugs.  This
    effectively provides self-service identities, as opposed to being provided externally.
    """

    _plugs = immp.ConfigProperty([immp.Plug])

    def __init__(self, name, config, host):
        super().__init__(name, _Schema.config(config), host)
        self.db = None

    async def start(self):
        if not self.config["instance"]:
            # Find a non-conflicting number and assign it.
            codes = {hook.config["instance"] for hook in self.host.hooks.values()
                     if isinstance(hook, self.__class__)}
            code = 1
            while code in codes:
                code += 1
            log.debug("Assigning instance code %d to hook %r", code, self.name)
            self.config["instance"] = code
        self.db = self.host.resources[DatabaseHook].db
        self.db.create_tables([IdentityGroup, IdentityLink, IdentityRole], safe=True)

    def get(self, name):
        """
        Retrieve the identity group using the given name.

        Args:
            name (str):
                Existing name to query.

        Returns:
            .IdentityGroup:
                Linked identity, or ``None`` if not linked.
        """
        try:
            return IdentityGroup.select_links().where(IdentityGroup.name == name).get()
        except IdentityGroup.DoesNotExist:
            return None

    def find(self, user):
        """
        Retrieve the identity that contains the given user, if one exists.

        Args:
            user (.User):
                Existing user to query.

        Returns:
            .IdentityGroup:
                Linked identity, or ``None`` if not linked.
        """
        if not user or user.plug not in self._plugs:
            return None
        try:
            return (IdentityGroup.select_links()
                                 .where(IdentityGroup.instance == self.config["instance"],
                                        IdentityLink.network == user.plug.network_id,
                                        IdentityLink.user == user.id).get())
        except IdentityGroup.DoesNotExist:
            return None

    async def channel_access(self, channel, user):
        return bool(IdentityLink.get(network=user.plug.network_id, user=user.id))

    async def identity_from_name(self, name):
        group = self.get(name)
        return await group.to_identity(self.host) if group else None

    async def identity_from_user(self, user):
        group = self.find(user)
        return await group.to_identity(self.host) if group else None

    def _test(self, channel, user):
        return channel.plug in self._plugs

    @command("id-add", scope=CommandScope.private, test=_test)
    async def add(self, msg, name, pwd):
        """
        Create a new identity, or link to an existing one from a second user.
        """
        if not msg.user or msg.user.plug not in self._plugs:
            return
        if self.find(msg.user):
            text = "{} Already identified".format(CROSS)
        else:
            pwd = IdentityGroup.hash(pwd)
            exists = False
            try:
                group = IdentityGroup.get(instance=self.config["instance"], name=name)
                exists = True
            except IdentityGroup.DoesNotExist:
                group = IdentityGroup.create(instance=self.config["instance"], name=name, pwd=pwd)
            if exists and not group.pwd == pwd:
                text = "{} Password incorrect".format(CROSS)
            elif not self.config["multiple"] and any(link.network == msg.user.plug.network_id
                                                     for link in group.links):
                text = "{} Already identified on {}".format(CROSS, msg.user.plug.network_name)
            else:
                IdentityLink.create(group=group, network=msg.user.plug.network_id,
                                    user=msg.user.id)
                text = "{} {}".format(TICK, "Added" if exists else "Claimed")
        await msg.channel.send(immp.Message(text=text))

    @command("id-rename", scope=CommandScope.private, test=_test)
    async def rename(self, msg, name):
        """
        Rename the current identity.
        """
        if not msg.user:
            return
        group = self.find(msg.user)
        if not group:
            text = "{} Not identified".format(CROSS)
        elif group.name == name:
            text = "{} No change".format(TICK)
        elif IdentityGroup.select().where(IdentityGroup.instance == self.config["instance"],
                                          IdentityGroup.name == name).exists():
            text = "{} Name already in use".format(CROSS)
        else:
            group.name = name
            group.save()
            text = "{} Claimed".format(TICK)
        await msg.channel.send(immp.Message(text=text))

    @command("id-password", scope=CommandScope.private, test=_test)
    async def password(self, msg, pwd):
        """
        Update the password for the current identity.
        """
        if not msg.user:
            return
        group = self.find(msg.user)
        if not group:
            text = "{} Not identified".format(CROSS)
        else:
            group.pwd = IdentityGroup.hash(pwd)
            group.save()
            text = "{} Changed".format(TICK)
        await msg.channel.send(immp.Message(text=text))

    @command("id-reset", scope=CommandScope.private, test=_test)
    async def reset(self, msg):
        """
        Delete the current identity and all linked users.
        """
        if not msg.user:
            return
        group = self.find(msg.user)
        if not group:
            text = "{} Not identified".format(CROSS)
        else:
            group.delete_instance()
            text = "{} Reset".format(TICK)
        await msg.channel.send(immp.Message(text=text))

    @command("id-role", scope=CommandScope.private, role=CommandRole.admin, test=_test)
    async def role(self, msg, name, role=None):
        """
        List roles assigned to an identity, or add/remove a given role.
        """
        try:
            group = IdentityGroup.get(instance=self.config["instance"], name=name)
        except IdentityGroup.DoesNotExist:
            text = "{} Name not registered".format(CROSS)
        else:
            if role:
                count = IdentityRole.delete().where(IdentityRole.group == group,
                                                    IdentityRole.role == role).execute()
                if count:
                    text = "{} Removed".format(TICK)
                else:
                    IdentityRole.create(group=group, role=role)
                    text = "{} Added".format(TICK)
            else:
                roles = IdentityRole.select().where(IdentityRole.group == group)
                if roles:
                    labels = [role.role for role in roles]
                    text = "Roles for {}: {}".format(name, ", ".join(labels))
                else:
                    text = "No roles for {}.".format(name)
        await msg.channel.send(immp.Message(text=text))



import os.path

from setuptools import setup


README = os.path.join(os.path.abspath(os.path.dirname(__file__)), "README.rst")

setup(name="IMMP",
      version="0.6.0.r1",
      author="Terrance",
      author_email="immp@terrance.allofti.me",
      url="https://immp.t.allofti.me",
      description="A modular processing platform for instant messages.",
      long_description=open(README).read(),
      long_description_content_type="text/x-rst",
      license="BSD 3-Clause License",
      platforms=["Any"],
      packages=["immp", "immp.core", "immp.hook", "immp.hook.webui", "immp.plug"],
      package_data={"immp.hook.webui": ["templates/**"]},
      entry_points={"console_scripts": ["immp=immp.__main__:entrypoint"]},
      python_requires=">=3.6",
      install_requires=["aiohttp>=3.0.0",
                        "voluptuous>=0.11.3"],
      extras_require={"runner": ["anyconfig>=0.9.5",
                                 "ruamel.yaml>=0.15.75"],
                      "db": ["peewee>=3.0.0"],
                      "web": ["aiohttp_jinja2>=1.0.0"],
                      "console": ["aioconsole>=0.1.14",
                                  "ptpython>=2.0.1"],
                      "sync": ["emoji>=0.5.0",
                               "jinja2>=2.6"],
                      "discord": ["discord.py>=1.0.0"],
                      "hangouts": ["hangups>=0.4.5"],
                      "telegram": ["telethon>=1.0"]},
      classifiers=["Development Status :: 4 - Beta",
                   "Intended Audience :: Developers",
                   "Topic :: Communications :: Chat",
                   "Topic :: Software Development :: Libraries"])



from argparse import ArgumentParser
import asyncio
import logging
import logging.config
import sys

import anyconfig
from voluptuous import ALLOW_EXTRA, REMOVE_EXTRA, Any, Optional, Schema

from immp import Channel, ConfigError, Group, Host, resolve_import
from immp.hook.runner import RunnerHook


class _Schema:

    _plugs = Any({str: {"path": str, Optional("config", default=dict): dict}}, {})

    _hooks = Any({str: {"path": str,
                        Optional("priority", default=None): Any(int, None),
                        Optional("config", default=dict): dict}}, {})

    _channels = Any({str: {"plug": str, "source": str}}, {})

    _logging = Schema({Optional("disable_existing_loggers", default=False): bool},
                      extra=ALLOW_EXTRA)

    config = Schema({Optional("path", default=list): [str],
                     Optional("plugs", default=dict): _plugs,
                     Optional("channels", default=dict): _channels,
                     Optional("groups", default=dict): Any({str: dict}, {}),
                     Optional("hooks", default=dict): _hooks,
                     Optional("logging", default=None): Any(_logging, None)},
                    extra=REMOVE_EXTRA, required=True)


class LocalFilter(logging.Filter):

    def filter(self, record):
        return record.name == "__main__" or record.name.split(".", 1)[0] == "immp"


def config_to_host(config, path, write):
    host = Host()
    for name, spec in config["plugs"].items():
        cls = resolve_import(spec["path"])
        host.add_plug(cls(name, spec["config"], host))
    for name, spec in config["channels"].items():
        plug = host.plugs[spec["plug"]]
        host.add_channel(name, Channel(plug, spec["source"]))
    for name, group in config["groups"].items():
        host.add_group(Group(name, group, host))
    for name, spec in config["hooks"].items():
        cls = resolve_import(spec["path"])
        host.add_hook(cls(name, spec["config"], host), spec["priority"])
    try:
        host.add_hook(RunnerHook("runner", {}, host))
    except ConfigError:
        # Prefer existing hook defined within the config itself.
        pass
    host.resources[RunnerHook].load(config, path, write)
    host.loaded()
    return host


def init(logs, paths):
    if logs:
        logging.config.dictConfig(logs)
    else:
        logging.basicConfig(level=logging.INFO)
        for handler in logging.root.handlers:
            handler.addFilter(LocalFilter())
    for search in paths:
        sys.path.append(search)


def main(path, write=False):
    config = _Schema.config(anyconfig.load(path))
    init(config["logging"], config["path"])
    log = logging.getLogger(__name__)
    log.info("Creating plugs and hooks")
    host = config_to_host(config, path, write)
    loop = asyncio.get_event_loop()
    task = loop.create_task(host.run())
    try:
        log.info("Starting host")
        loop.run_until_complete(task)
    except KeyboardInterrupt:
        log.info("Closing on interrupt")
        task.cancel()
        loop.run_until_complete(task)
    finally:
        loop.close()
        if write:
            host.resources[RunnerHook].write_config()


def entrypoint():
    parser = ArgumentParser(prog="python -m immp", add_help=False)
    parser.add_argument("-w", "--write", action="store_true")
    parser.add_argument("file", metavar="FILE")
    args = parser.parse_args()
    main(args.file, args.write)


if __name__ == "__main__":
    entrypoint()
